import collections
from glob import glob
import os, yaml
import numpy as np
from scipy import stats
from matplotlib import pyplot as plt
import pandas as pd

from matplotlib import rcParams

from xbbo.utils.util import loadJson, loadOBJ
from xbbo.utils.constants import Key


class Analyse():
    def __init__(self,
                 exp_dir_root='./exp',
                 benchmark='countingones',
                 methods=None,
                 marks=None,
                 limit=1e7,
                 **kwargs) -> None:
        self.exp_dir_root = exp_dir_root
        self.out_dir = os.path.join(exp_dir_root, benchmark)
        if not os.path.exists(self.out_dir):
            os.mkdir(self.out_dir)
        self.methods = methods
        self.marks = marks
        self.benchmark = benchmark
        self._set_plot()
        plt.clf()
        hashset = set()
        min_cost = np.inf
        max_cost = 0
        self.min_regret = np.inf
        self.max_regret = -np.inf
        self.regret_key = Key.REGRET_TEST
        # new first
        cfg_paths = sorted(glob(exp_dir_root + '/*/*.yaml'), reverse=True)
        mean_df = {}
        std_df = {}
        index = -1
        for cfg_path in cfg_paths:  # for every method
            with open(cfg_path, 'r') as f:
                cfg = yaml.safe_load(f)
            # hash_name = cfg_path.split('/')[-2].split('_')[-1]
            mark_label = cfg["mark_label"]
            method_name = cfg["OPTM"]["name"]  # tmp[-1][:-5]
            if cfg["TEST_PROBLEM"]["name"] != self.benchmark or (
                (self.methods is not None) and
                (method_name not in self.methods)) or (
                    (self.marks is not None) and
                    (mark_label not in self.marks)) or mark_label in hashset:
                continue
            jfiles = glob('{}/{}/*.json'.format(cfg["GENERAL"]["exp_dir"],
                                                method_name))
            if len(jfiles) == 0:
                continue
            hashset.add(mark_label)
            index += 1

            regret = []
            costs = []
            for jfile in jfiles:
                j = loadJson(jfile)
                curr_regret = np.array(j[self.regret_key])
                if method_name == 'dehb_':
                    curr_cost = np.array(j['runtime'])
                else:
                    curr_cost = np.array(j[Key.COST])
                if self.benchmark == "countingones":
                    d = cfg["TEST_PROBLEM"].get("n_continuous",
                                                4) + cfg["TEST_PROBLEM"].get(
                                                    "n_categorical", 4)
                    if method_name == 'dehb_':
                        curr_regret = curr_regret * d - d
                    curr_regret = (curr_regret + d) / d  # 0-1
                    max_budget = 93312 / d
                    curr_cost /= max_budget
                _, idx = np.unique(curr_regret, return_index=True)
                idx.sort()

                regret.append(curr_regret[idx])
                costs.append(curr_cost[idx])
            # regret = np.array(regret)
            # costs = np.array(costs)

            # finds the latest time where the first measurement was made across runs
            t = np.max([costs[i][0] for i in range(len(costs))])
            # t = costs[:, 0].max()
            min_cost = min(min_cost, t)
            te, cost = self._fill_trajectory(regret, costs, replace_nan=1)

            idx = cost.tolist().index(t)
            te = te[idx:, :]
            cost = cost[idx:]

            # Clips off all measurements after 10^7s
            idx = np.where(cost <= limit)[0]
            # if hash_to_label_map is not None:
            #     label_name = hash_to_label_map.get(hash_name, method_name+'-'+hash_name)
            # else:
            #     label_name = method_name+'-'+hash_name
            print("{}. Plotting for {}".format(index, mark_label))
            print(len(regret), len(costs))
            print("\nMean: {}; Std: {}\n".format(
                np.mean(te, axis=1)[idx][-1],
                stats.sem(te[idx], axis=1)[-1]))
            # The mean plot
            plt.plot(cost[idx],
                     np.mean(te, axis=1)[idx],
                     color='C{}'.format(index),
                     linewidth=4,
                     label=mark_label,
                     linestyle=self.linestyles[index % len(self.linestyles)],
                     marker=self.marker[index % len(self.marker)],
                     markevery=(0.1, 0.1),
                     markersize=15)
            # The error band
            plt.fill_between(
                cost[idx],
                np.mean(te[idx], axis=1)[idx] + 2 * stats.sem(te[idx], axis=1),
                np.mean(te[idx], axis=1)[idx] - 2 * stats.sem(te[idx], axis=1),
                color="C%d" % index,
                alpha=0.2)

            # Stats to dynamically impose limits on the axes of the plots
            max_cost = max(max_cost, cost[idx][-1])
            self.min_regret = min(self.min_regret,
                                  np.mean(te, axis=1)[idx][-1])
            self.max_regret = max(self.max_regret, np.mean(te, axis=1)[idx][0])

            # For final score table
            mean_df[mark_label] = pd.Series(data=np.mean(te, axis=1)[idx],
                                            index=cost[idx])
            std_df[mark_label] = pd.Series(data=np.std(te, axis=1)[idx],
                                           index=cost[idx])
        mean_df = pd.DataFrame(mean_df)
        all_mean_df = mean_df.copy()
        all_mean_df.ffill().to_pickle(
            os.path.join(self.out_dir, 'all_mean_df.pkl'))
        std_df = pd.DataFrame(std_df)
        # minimum of the maximum time limit recorded for each algorithm
        cutoff_idx = min(
            list(
                map(lambda x: np.where(~mean_df.isna()[x] == True)[0][-1],
                    mean_df.columns)))
        mean_df = mean_df.iloc[:cutoff_idx + 1].ffill()
        std_df = std_df.iloc[:cutoff_idx + 1].ffill()
        if len(hashset) > 1:
            rank_df = mean_df.apply(stats.rankdata,
                                    axis=1,
                                    result_type='broadcast')
            rank_df.to_pickle(os.path.join(self.out_dir, 'rank_df.pkl'))
        mean_df.iloc[-1].to_pickle(os.path.join(self.out_dir, 'mean_df.pkl'))
        std_df.iloc[-1].to_pickle(os.path.join(self.out_dir, 'std_df.pkl'))

        # self.plt = plt
        self.min_cost = min_cost
        self.max_cost = max_cost
        # self.min_regret = self.min_regret
        # self.max_regret = self.max_regret
        self._regret_plot(**kwargs)

    def _set_plot(self, fix_colors=False):
        rcParams["font.size"] = "25"
        rcParams['text.usetex'] = False
        rcParams['font.family'] = 'serif'
        rcParams['figure.figsize'] = (16.0, 9.0)
        rcParams['figure.frameon'] = True
        rcParams['figure.edgecolor'] = 'k'
        rcParams['grid.color'] = 'k'
        rcParams['grid.linestyle'] = ':'
        rcParams['grid.linewidth'] = 0.5
        rcParams['axes.linewidth'] = 1
        rcParams['axes.edgecolor'] = 'k'
        rcParams['axes.grid.which'] = 'both'
        rcParams['legend.frameon'] = 'True'
        rcParams['legend.framealpha'] = 1

        rcParams['ytick.major.size'] = 12
        rcParams['ytick.major.width'] = 1.5
        rcParams['ytick.minor.size'] = 6
        rcParams['ytick.minor.width'] = 1
        rcParams['xtick.major.size'] = 12
        rcParams['xtick.major.width'] = 1.5
        rcParams['xtick.minor.size'] = 6
        rcParams['xtick.minor.width'] = 1
        self.marker = ['x', '^', 'D', 'o', 's', 'h', '*', 'v', '<', ">"]
        self.linestyles = ['-', '--', '-.', ':']
        # plot setup
        # self.colors = ["C%d" % i for i in range(len(self.methods))]
        # if fix_colors and len(self.methods) <= 8:
        #     _colors = dict()
        #     _colors["RS"] = "C0"
        #     _colors["HB"] = "C7"
        #     _colors["BOHB"] = "C1"
        #     _colors["TPE"] = "C3"
        #     _colors["SMAC"] = "C4"
        #     _colors["RE"] = "C5"
        #     _colors["DE"] = "C6"
        #     _colors["DEHB"] = "C2"
        #     self.colors = []
        #     for (_, l) in self.methods:
        #         self.colors.append(_colors[l])

    def _fill_trajectory(self,
                         performance_list,
                         cost_list,
                         replace_nan=np.NaN):
        frame_dict = collections.OrderedDict()
        counter = np.arange(0, len(performance_list))
        for p, t, c in zip(performance_list, cost_list, counter):
            if len(p) != len(t):
                raise ValueError("(%d) Array length mismatch: %d != %d" %
                                 (c, len(p), len(t)))
            frame_dict[str(c)] = pd.Series(data=p, index=t)

        # creates a dataframe where the rows are indexed based on time
        # fills with NA for missing values for the respective timesteps
        merged = pd.DataFrame(frame_dict)
        # ffill() acts like a fillna() wherein a forward fill happens
        # only remaining NAs for in the beginning until a value is recorded
        merged = merged.ffill()

        performance = merged.to_numpy()  # converts to a 2D numpy array
        cost_ = merged.index.values  # retrieves the timestamps

        performance[np.isnan(performance)] = replace_nan

        if not np.isfinite(performance).all():
            raise ValueError(
                "\nCould not merge lists, because \n"
                "\t(a) one list is empty?\n"
                "\t(b) the lists do not start with the same times and"
                " replace_nan is not set?\n"
                "\t(c) any other reason.")

        return performance, cost_

    def _regret_plot(self, **kwargs):
        plot_name = kwargs.get("plot_name", "comparison")
        output_type = kwargs.get("output_type", "pdf")
        plot_type = kwargs.get("plot_type", "wallclock")
        # bench_type = kwargs.get("bench_type", 'protein')
        title = kwargs.get("title", "benchmark")
        legend_size = kwargs.get("legend_size", 40)
        if self.benchmark != 'cc18':
            plt.xscale("log")
        if self.benchmark != 'svm' and self.benchmark != 'bnn':
            plt.yscale("log")
        plt.tick_params(which='both', direction="in")
        if self.benchmark == 'svm' or self.benchmark == 'bnn' or self.benchmark == "cc18" or self.benchmark == "paramnet":
            plt.legend(loc='upper right',
                       framealpha=1,
                       prop={
                           'size': legend_size,
                           'weight': 'normal'
                       })
        elif self.benchmark == "rl":
            plt.legend(loc='lower left',
                       framealpha=1,
                       prop={
                           'size': legend_size,
                           'weight': 'normal'
                       },
                       ncol=1)
        else:  #elif self.benchmark == "countingones":
            plt.legend(loc='lower left',
                       framealpha=1,
                       prop={
                           'size': legend_size,
                           'weight': 'normal'
                       })
        plt.title(title, size=40)

        if self.benchmark == 'rl':
            plt.xlabel("time $[s]$", fontsize=45)
        elif self.benchmark == 'bnn':
            plt.xlabel("MCMC steps", fontsize=45)
        elif self.benchmark == 'countingones':
            plt.xlabel("cummulative budget / $b_{max}$", fontsize=45)
        elif self.benchmark == 'speed':
            plt.xlabel("Runtime sans function evalution")
        elif plot_type == "wallclock":
            plt.xlabel("estimated wallclock time $[s]$", fontsize=45)
        elif plot_type == "fevals":
            plt.xlabel("number of function evaluations", fontsize=45)

        if self.benchmark == 'svm':
            plt.ylabel("{} error".format(self.regret_key), fontsize=45)
        elif self.benchmark == 'rl':
            plt.ylabel("epochs until convergence", fontsize=45)
        elif self.benchmark == 'bnn':
            plt.ylabel("negative log-likelihood", fontsize=45)
        elif self.benchmark == 'countingones':
            plt.ylabel("normalized {} regret".format(self.regret_key),
                       fontsize=40)
        elif self.benchmark == 'countingones':
            plt.ylabel("number of function evaluations", fontsize=45)
        else:
            plt.ylabel("{} regret".format(self.regret_key), fontsize=45)

        if self.benchmark == 'rl':
            # plt.xlim(1e2, 1e5)
            plt.xlim(1e2, self.max_cost)
            # plt.xlim(self.min_cost, self.max_cost)
        elif self.benchmark == 'bnn':
            # plt.xlim(min_limit, self.max_cost)
            plt.xlim(50000, self.max_cost)  # min(self.max_cost*10, limit))
        elif self.benchmark == 'countingones':
            plt.xlim(self.min_cost, self.max_cost)
            # plt.xlim(self.min_cost, 1e4)
        elif self.benchmark == 'cc18':
            plt.xlim(0.1, self.max_cost)
        elif self.benchmark == "paramnet":
            print("Max time: {}".format(self.max_cost))
            plt.xlim(self.min_cost, self.max_cost)
        elif self.benchmark == "nas_101_cifar10":
            plt.xlim(1e2, self.max_cost)
        else:
            plt.xlim(self.min_cost, self.max_cost)
            # plt.xlim(max(self.min_cost/10, 1e0), min(self.max_cost*10, 1e7))

        if self.benchmark == 'bnn':
            plt.ylim(3, 10)  # 75)
        elif self.benchmark == 'rl':
            plt.ylim(1e2, 1e4)
        elif self.benchmark == 'cc18':
            plt.ylim(0, self.max_regret)
        elif self.benchmark == 'svm':
            plt.ylim(self.min_regret, 0.5)
        else:
            plt.ylim(self.min_regret, self.max_regret)

        plt.grid(which='both', alpha=0.2, linewidth=0.5)
        print(
            os.path.join(self.out_dir, '{}.{}'.format(plot_name, output_type)))
        plt.savefig(os.path.join(self.out_dir,
                                 '{}.{}'.format(plot_name, output_type)),
                    bbox_inches='tight')


class Analyse_multi_benchmark():
    def __init__(self, exp_dir_root='./exp', **kwargs) -> None:
        self.exp_dir_root = exp_dir_root
        list_of_mean_files = glob(
            os.path.join(self.exp_dir_root, '*/mean_df.pkl'))

        list_of_std_files = glob(
            os.path.join(self.exp_dir_root, '*/std_df.pkl'))
        self.mean_dfs = {}
        for filename in list_of_mean_files:
            benchname = filename.split('/')[-2]
            self.mean_dfs[benchname] = loadOBJ(filename)
        self.mean_dfs = pd.DataFrame(self.mean_dfs).transpose()
        self.mean_dfs.to_pickle(
            os.path.join(self.exp_dir_root, "all_mean_dfs.pkl"))

        std_dfs = {}
        for filename in list_of_std_files:
            benchname = filename.split('/')[-2]
            std_dfs[benchname] = loadOBJ(filename)
        std_dfs = pd.DataFrame(std_dfs).transpose()
        std_dfs.to_pickle(os.path.join(self.exp_dir_root, "all_std_dfs.pkl"))

        # Load run statistics to create a relative ranking plot over time

        rank_list_candidates = glob(
            os.path.join(self.exp_dir_root, '*/rank_df.pkl'))
        list_of_rank_files = []
        for name in rank_list_candidates:
            # ignore benchmarks where the runtime is not wallclock time in seconds
            # if "countingones" in name or "bnn" in name or "svm" in name:
            #     continue
            list_of_rank_files.append(name)

        # load rankings per benchmark
        rank_dfs = []
        for filename in list_of_rank_files:
            rank_dfs.append(loadOBJ(filename))
        # reorganize data to have algorithms as the top hierarchy, followed by every benchmark for the algo
        avg_rank = {}
        for i in range(len(rank_dfs)):
            for name in rank_dfs[i].columns:
                if name not in avg_rank.keys():
                    avg_rank[name] = {}
                if i not in avg_rank[name].keys():
                    avg_rank[name][i] = None
                avg_rank[name][i] = pd.Series(data=rank_dfs[i][name],
                                              index=rank_dfs[i].index)

        # assigning mean rank to all algorithms at start
        starting_rank = np.mean(np.arange(1, 1 + len(avg_rank.keys())))
        for name, v in avg_rank.items():
            avg_rank[name] = pd.DataFrame(v)
            avg_rank[name].iloc[0] = [starting_rank] * avg_rank[name].shape[1]

        # compute mean relative rank of each algorithm across all benchmarks
        self.rank_lists = {}
        for name, v in avg_rank.items():
            self.rank_lists[name] = pd.Series(data=np.mean(
                avg_rank[name].ffill(), axis=1),
                                              index=avg_rank[name].index)
        self.rank_lists = pd.DataFrame(self.rank_lists)

        self.linestyles = [
            (0, (1, 5)),  # loosely dotted
            (0, (5, 5)),  # loosely dashed
            'dotted',
            (0, (3, 2, 1, 2, 1, 2)),  # dash dot dotted
            'dashed',
            'dashdot',
            (0, (3, 1, 1, 1, 1, 1)),
            'solid'
        ]

        self.colors = ["C%d" % i for i in range(len(self.rank_lists.columns))]
        # if len(rank_lists.columns) <= 8:
        #     _colors = dict()
        #     _colors["RS"] = "C0"
        #     _colors["HB"] = "C7"
        #     _colors["BOHB"] = "C1"
        #     _colors["TPE"] = "C3"
        #     _colors["SMAC"] = "C4"
        #     _colors["RE"] = "C5"
        #     _colors["DE"] = "C6"
        #     _colors["DEHB"] = "C2"
        #     colors = []
        #     for l in rank_lists.columns:
        #         colors.append(_colors[l])

        self.landmarks = np.arange(start=0,
                                   stop=self.rank_lists.shape[0],
                                   step=5)  # for smoothing
        self._set_plot()
        self.rank_plot()

    def _set_plot(self, ):
        rcParams['font.family'] = 'serif'

    def rank_plot(self, ):
        plt.clf()
        xlims = [np.inf, -np.inf]
        for i, name in enumerate(self.rank_lists.columns):
            if name == 'DEHB':
                lw, a = (1.75, 1)
            else:
                lw, a = (1.5, 0.7)
            plt.plot(self.rank_lists[name].index.to_numpy()[self.landmarks],
                     self.rank_lists[name].to_numpy()[self.landmarks],
                     label=name,
                     alpha=a,
                     linestyle=self.linestyles[i],
                     linewidth=1.5,
                     color=self.colors[i])
            xlims[0] = min(xlims[0], self.rank_lists[name].index.to_numpy()[0])
            xlims[1] = max(xlims[1],
                           self.rank_lists[name].index.to_numpy()[-1])

        plt.xscale('log')
        plt.legend(loc='upper left', framealpha=1, prop={'size': 12}, ncol=4)
        # plt.fill_between(
        #     rank_lists['DEHB'].index.to_numpy()[landmarks],
        #     0, rank_lists['DEHB'].to_numpy()[landmarks],
        #     alpha=0.5, color=_colors["DEHB"]
        # )
        # plt.fill_between(
        #     rank_lists['DEHB'].index.to_numpy()[landmarks],
        #     0, starting_rank,
        #     alpha=0.3, color='gray'
        # )
        # plt.hlines(starting_rank, 0, 1e7)
        plt.xlim(xlims[0], xlims[1])
        plt.ylim(1, self.rank_lists.shape[1])
        plt.xlabel('estimated wallclock time $[s]$', fontsize=15)
        plt.ylabel('average relative rank', fontsize=15)
        plt.savefig(os.path.join(self.exp_dir_root, 'rank_plot.pdf'),
                    bbox_inches='tight')

        rank_stats = {}
        rank_stats['minimum'] = np.min(self.rank_lists, axis=0)
        rank_stats['maximum'] = np.max(self.rank_lists, axis=0)
        rank_stats['variance'] = np.var(self.rank_lists, axis=0)
        rank_stats = pd.DataFrame(rank_stats)
        # Ranks based on final numbers
        rank_df = {}
        for idx in self.mean_dfs.index:
            rank_df[idx] = pd.Series(data=stats.rankdata(
                self.mean_dfs.loc[idx]),
                                     index=self.mean_dfs.loc[idx].index)
        rank_df = pd.DataFrame(rank_df)
        print(rank_df.mean(axis=1))


import numpy as np
from time import time
import tqdm
from matplotlib import pyplot as plt

from xbbo.search_space import problem_register
from xbbo.search_algorithm import alg_register
from xbbo.configspace import build_space


class PBT:

    def __init__(self, cfg):
        # setup TestProblem
        self.cfg = cfg
        self.pop_size = cfg.OPTM.pop_size

        self.population_model = [problem_register[cfg.TEST_PROBLEM.name](cfg) for _ in range(self.pop_size)]
        self.api_config = self.population_model[0].get_api_config()  # 优化的hp
        self.config_spaces = build_space(self.api_config)

        # Setup optimizer
        opt_class = alg_register[cfg.OPTM.name]

        self.optimizer_instance = opt_class(self.config_spaces, self.pop_size, **dict(cfg.OPTM.kwargs))
        self.n_suggestions = cfg.OPTM.n_suggestions
        self.n_obj = cfg.OPTM.n_obj

        assert self.n_suggestions >= 1, "batch size must be at least 1"
        assert self.n_obj >= 1, "Must be at least one objective"

        self.epoch = cfg.OPTM.epoch
        self.interval = cfg.OPTM.interval

        # self.record = Record(self.cfg)

    def evaluate(self, population_model_history_hp):
        model = self.population_model[0]
        for step_num, params, acc in population_model_history_hp:
            model.update_hp(params)
            for i in range(step_num):
                model.step()

    def run(self):
        self.optimizer_instance.init_model_hp(self.population_model)
        finished = False
        with tqdm.tqdm(total=int(len(self.population_model[-1]) * self.epoch)) as pbar:
            while not finished:
                for i in range(self.pop_size):
                    self.population_model[i].step(int(self.interval * len(self.population_model[i])))
                    if self.population_model[i].step_num == int(len(self.population_model[i]) * self.epoch):
                        finished = True
                    # while True:
                    #     self.population_model[i].step()
                    #     if self.population_model[i].step_num % (self.interval * len(self.population_model[i])) == 0:
                    #         if self.population_model[i].step_num == len(self.population_model[i]) * self.epoch:
                    #             finished = True
                    #         # self.population_model[i].ready = True
                    #         break
                # asynchronous wait all active
                for i in range(self.pop_size):
                    self.population_model[i].evaluate()
                scores = [net.score for net in self.population_model]
                pbar.update(self.interval * len(self.population_model[-1]))
                if finished:
                    break
                self.optimizer_instance.exploit_and_explore(self.population_model, scores)
                # self.optimizer_instance.exploit_and_explore_toy(self.population_model, scores)
        return scores

    def show_res(self, scores):
        best_individual_index = np.argmax(scores)
        fig, (ax1, ax2) = plt.subplots(1, 2)
        for i in range(self.pop_size):
            desc_data = np.array(self.population_model[i].history_score)
            desc_data[:, 0] /= len(self.population_model[-1])
            ax1.plot(desc_data[:, 0], desc_data[:, 1], alpha=0.5)
        ax1.set_xlabel("epoch")
        ax1.set_ylabel("score")
        # for i in range(self.pop_size):
        #     desc_data = np.array([list(x[-1].values()) for x in self.population_model[i].trajectory_hp])
        #     # desc_data[:, 0] /= self.interval * len(self.population_model[-1])
        #     ax2.scatter(desc_data[:, 0], desc_data[:, 1], alpha=0.5)
        # ax2.set_xlabel("hp_1")
        # ax2.set_ylabel("hp_2")
        for i in range(self.pop_size):
            desc_data = np.array([[x[0], x[-1]['lr']] for x in self.population_model[i].history_hp])
            desc_data[:, 0] /= len(self.population_model[-1])
            desc_data = np.append(desc_data, [[self.epoch, desc_data[-1, 1]]], axis=0)
            ax2.plot(desc_data[:, 0], desc_data[:, 1], label='best individual' if i==best_individual_index else None)
        ax2.set_xlabel("epoch")
        ax2.set_ylabel("lr")
        plt.legend()
        plt.suptitle("PBT search (lr, momentum) in MNIST")
        plt.tight_layout()
        plt.savefig('./out/PBT_mnist.png')
        plt.show()

        print('-----\nBest hyper-param strategy: {}'.format(self.population_model[best_individual_index].history_hp))
        print('final score: {}'.format(self.population_model[best_individual_index].history_score[-1]))

    def show_toy_res(self, scores):
        best_individual_index = np.argmax(scores)
        fig, (ax1, ax2) = plt.subplots(1, 2)
        for i in range(self.pop_size):
            desc_data = np.array(self.population_model[i].history_score)
            desc_data[:, 0] /= len(self.population_model[-1])
            ax1.plot(desc_data[:, 0], desc_data[:, 1], alpha=0.5)
        ax1.set_xlabel("epoch")
        ax1.set_ylabel("score")
        for i in range(self.pop_size):
            desc_data = np.array(self.population_model[i].trajectory_theta)
            # desc_data[:, 0] /= self.interval * len(self.population_model[-1])
            ax2.scatter(desc_data[:, 0], desc_data[:, 1], s=2, alpha=0.5)
        # ax2.axis('equal')
        ax2.set_xlim(0, 1)
        ax2.set_ylim(0, 1)
        ax2.set_xlabel(r"$\theta_1$")
        ax2.set_ylabel(r"$\theta_2$")
        # for i in range(self.pop_size):
        #     desc_data = np.array([[x[0], x[-1]['lr']] for x in self.population_model[i].history_hp])
        #     desc_data[:, 0] /= len(self.population_model[-1])
        #     desc_data = np.append(desc_data, [[self.epoch, desc_data[-1, 1]]], axis=0)
        #     ax2.plot(desc_data[:, 0], desc_data[:, 1], label='best individual' if i==best_individual_index else None)
        # ax2.set_xlabel("epoch")
        # ax2.set_ylabel("lr")
        # plt.legend()
        plt.suptitle("PBT toy example")
        plt.tight_layout()
        plt.savefig('./out/PBT_toy.png')
        plt.show()

        print('-----\nBest hyper-param strategy: {}'.format(self.population_model[best_individual_index].history_hp))
        print('final score: {}'.format(self.population_model[best_individual_index].history_score[-1]))



'''
Reference: https://github.com/automl/DEHB
'''

import logging
from typing import List
from sklearn.model_selection import KFold
import numpy as np
from xbbo.acquisition_function.acq_func import EI_AcqFunc
from xbbo.acquisition_function.acq_optimizer import InterleavedLocalAndRandomSearch, LocalSearch, RandomScipyOptimizer, RandomSearch, ScipyGlobalOptimizer, ScipyOptimizer
from xbbo.initial_design import ALL_avaliable_design
from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.search_algorithm.multi_fidelity.BOHB import BOHB

from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trials, Trial
from xbbo.search_algorithm.multi_fidelity.utils.bracket_manager import BasicConfigGenerator
try:
    from xbbo.surrogate.transfer.rf_with_instances import RandomForestWithInstances
except:
    print("If your want to use probability random foreset, make sure 'pyrfr' is installed!")
from xbbo.utils.constants import MAXINT, Key
from xbbo.utils.util import get_types
from .. import alg_register
logger = logging.getLogger(__name__)

def std_normalization(x):
    _mean = np.mean(x)
    _std = np.std(x)
    if _std == 0:
        return np.array([0.] * len(x))
    return (np.array(x) - _mean) / _std

@alg_register.register('mfes-bo')
class SMBO(AbstractOptimizer):
    def __init__(self,
                space: DenseConfigurationSpace,
                all_budgets: List,
                seed: int = 42,
                initial_design: str = 'random',
                suggest_limit: int = np.inf,
                surrogate: str = 'rf',
                # acq_func: str = 'ei',
                acq_opt: str = 'rs_ls',
                # predict_x_best: bool = False,
                weight_srategy: str = 'rank_loss_p_norm',
                fusion_method='idp',
                init_weight=None,
                update_enable=True,
                random_fraction=1 / 3,
                **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   seed=seed,
                                   suggest_limit=suggest_limit,
                                   **kwargs)
        self.logger = logger
        self.all_budgets = all_budgets
        self.update_enable = update_enable
        self.fusion_method = fusion_method
        self.dimension = self.space.get_dimensions()
        self.random_fraction = random_fraction
        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, ta_run_limit=suggest_limit, **kwargs)
        self.init_budget = self.initial_design.init_budget

        self.types, self.bounds = get_types(self.space)

        self.initial_design_configs = self.initial_design.select_configurations(
        )
        if init_weight is None:
            k = len(all_budgets)
            init_weight = [1. / (k-1)] * (k-1) + [0.]
        self.logger.info("Initialize weight to %s" %
                         init_weight)    
        self.trials = Trials(space,dim=self.dimension)
        if surrogate == 'rf':
            from xbbo.surrogate.transfer.rf_ensemble import RandomForestEnsemble
            self.weighted_surrogate = RandomForestEnsemble(space, all_budgets, init_weight, fusion_method,types=self.types, bounds=self.bounds,**kwargs
        )
        self.weight_srategy = weight_srategy
        if weight_srategy == 'rank_loss_p_norm':
                self.power_num = kwargs.get("power_num", 3)
        self.acquisition_func = EI_AcqFunc(self.weighted_surrogate, self.rng)    
        if acq_opt == 'ls':
            self.acq_maximizer = LocalSearch(self.acquisition_func, self.space,
                                             self.rng)
        elif acq_opt == 'rs':
            self.acq_maximizer = RandomSearch(self.acquisition_func,
                                              self.space, self.rng)
        elif acq_opt == 'rs_ls':
            self.acq_maximizer = InterleavedLocalAndRandomSearch(
                self.acquisition_func, self.space, self.rng)
        elif acq_opt == 'scipy':
            self.acq_maximizer = ScipyOptimizer(self.acquisition_func,
                                                self.space, self.rng)
        elif acq_opt == 'scipy_global':
            self.acq_maximizer = ScipyGlobalOptimizer(self.acquisition_func,
                                                      self.space, self.rng)
        elif acq_opt == 'r_scipy':
            self.acq_maximizer = RandomScipyOptimizer(self.acquisition_func,
                                                      self.space, self.rng)
        else:
            raise ValueError('acq_opt {} not in {}'.format(
                acq_opt,
                ['ls', 'rs', 'rs_ls', 'scipy', 'scipy_global', 'r_scipy']))
        self.weight_changed_cnt = 0
        self.target_x = dict()
        self.target_y = dict()
        for budget in self.all_budgets:
            self.target_x[budget] = []
            self.target_y[budget] = []

    @staticmethod
    def calculate_preserving_order_num(y_pred, y_true):
        array_size = len(y_pred)
        assert len(y_true) == array_size

        total_pair_num, order_preserving_num = 0, 0
        for idx in range(array_size):
            for inner_idx in range(idx + 1, array_size):
                if bool(y_true[idx] > y_true[inner_idx]) == bool(y_pred[idx] > y_pred[inner_idx]):
                    order_preserving_num += 1
                total_pair_num += 1
        return order_preserving_num, total_pair_num

    def update_weight(self):

        max_budget = self.all_budgets[-1]
        incumbent_configs = self.target_x[max_budget]
        if not self.target_x[max_budget]:
            return 
        test_x = np.array(incumbent_configs)
        test_y = np.array(self.target_y[max_budget], dtype=np.float64)

        # budget_list = self.weighted_surrogate.surrogate_budget
        budget_list = self.all_budgets
        K = len(budget_list)

        if len(test_y) >= 3:
            # Get previous weights
            if self.weight_srategy == 'rank_loss_p_norm':
                preserving_order_p = list()
                preserving_order_nums = list()
                for i, r in enumerate(budget_list):
                    fold_num = 5
                    if i != K - 1:
                        mean, var = self.weighted_surrogate.surrogate_container[r].predict(test_x)
                        tmp_y = np.reshape(mean, -1)
                        preorder_num, pair_num = self.calculate_preserving_order_num(tmp_y, test_y)
                        preserving_order_p.append(preorder_num / pair_num)
                        preserving_order_nums.append(preorder_num)
                    else:
                        if len(test_y) < 2 * fold_num:
                            preserving_order_p.append(0)
                        else:
                            # 5-fold cross validation.
                            kfold = KFold(n_splits=fold_num)
                            cv_pred = np.array([0] * len(test_y))
                            for train_idx, valid_idx in kfold.split(test_x):
                                train_configs, train_y = test_x[train_idx], test_y[train_idx]
                                valid_configs, valid_y = test_x[valid_idx], test_y[valid_idx]
                                _surrogate = RandomForestWithInstances(types=self.types, bounds=self.bounds)
                                _surrogate.train(train_configs, train_y)
                                pred, _ = _surrogate.predict(valid_configs)
                                cv_pred[valid_idx] = pred.reshape(-1)
                            preorder_num, pair_num = self.calculate_preserving_order_num(cv_pred, test_y)
                            preserving_order_p.append(preorder_num / pair_num)
                            preserving_order_nums.append(preorder_num)

                trans_order_weight = np.array(preserving_order_p)
                power_sum = np.sum(np.power(trans_order_weight, self.power_num))
                if power_sum < 1e-9:
                    old_weights = list()
                    for i, r in enumerate(budget_list):
                        _weight = self.weighted_surrogate.surrogate_weight[r]
                        old_weights.append(_weight)
                    new_weights = old_weights.copy()
                else:
                    new_weights = np.power(trans_order_weight, self.power_num) / power_sum

            elif self.weight_srategy == 'rank_loss_prob':
                # For basic surrogate i=1:K-1.
                mean_list, var_list = list(), list()
                for i, r in enumerate(budget_list[:-1]):
                    mean, var = self.weighted_surrogate.surrogate_container[r].predict(test_x)
                    mean_list.append(np.reshape(mean, -1))
                    var_list.append(np.reshape(var, -1))
                sample_num = 100
                min_probability_array = [0] * K
                for _ in range(sample_num):
                    order_preseving_nums = list()

                    # For basic surrogate i=1:K-1.
                    for idx in range(K - 1):
                        sampled_y = self.rng.normal(mean_list[idx], var_list[idx])
                        _num, _ = self.calculate_preserving_order_num(sampled_y, test_y)
                        order_preseving_nums.append(_num)

                    fold_num = 5
                    # For basic surrogate i=K. cv
                    if len(test_y) < 2 * fold_num:
                        order_preseving_nums.append(0)
                    else:
                        # 5-fold cross validation.
                        kfold = KFold(n_splits=fold_num)
                        cv_pred = np.array([0] * len(test_y))
                        for train_idx, valid_idx in kfold.split(test_x):
                            train_configs, train_y = test_x[train_idx], test_y[train_idx]
                            valid_configs, valid_y = test_x[valid_idx], test_y[valid_idx]
                            _surrogate = RandomForestWithInstances(types=self.types, bounds=self.bounds)
                            _surrogate.train(train_configs, train_y)
                            _pred, _var = _surrogate.predict(valid_configs)
                            sampled_pred = self.rng.normal(_pred.reshape(-1), _var.reshape(-1))
                            cv_pred[valid_idx] = sampled_pred
                        _num, _ = self.calculate_preserving_order_num(cv_pred, test_y)
                        order_preseving_nums.append(_num)
                    max_id = np.argmax(order_preseving_nums)
                    min_probability_array[max_id] += 1
                new_weights = np.array(min_probability_array) / sample_num
            else:
                raise ValueError('Invalid weight method: %s!' % self.weight_srategy)
        else:
            old_weights = list()
            for i, r in enumerate(budget_list):
                _weight = self.weighted_surrogate.surrogate_weight[r]
                old_weights.append(_weight)
            new_weights = old_weights.copy()

        self.logger.info('[%s] %d-th Updating weights: %s' % (
            self.weight_srategy, self.weight_changed_cnt, str(new_weights)))

        # Assign the weight to each basic surrogate.
        for i, r in enumerate(budget_list):
            self.weighted_surrogate.surrogate_weight[r] = new_weights[i]
        self.weight_changed_cnt += 1
        # Save the weight data.
        # self.hist_weights.append(new_weights)
        # dir_path = os.path.join(self.data_directory, 'saved_weights')
        # file_name = 'mfes_weights_%s.npy' % (self.method_name,)
        # if not os.path.exists(dir_path):
        #     os.makedirs(dir_path)
        # np.save(os.path.join(dir_path, file_name), np.asarray(self.hist_weights))
        # self.logger.info('update_weight() cost %.2fs. new weights are saved to %s'
        #                  % (time.time()-start_time, os.path.join(dir_path, file_name)))
    def _suggest(self, n_suggestions=1):
        trial_list = []
        trials_budgets = np.array([float(info[Key.BUDGET]) for info in self.trials.infos])

        # currently only suggest one
        if (self.trials.trials_num) < self.init_budget :
            assert self.trials.trials_num % n_suggestions == 0
            configs = self.initial_design_configs[
                int(n_suggestions *
                    self.trials.trials_num):int(n_suggestions *
                                                (self.trials.trials_num + 1))]
            for config in configs:
                trial_list.append(
                    Trial(configuration=config,
                          config_dict=config.get_dictionary(),
                          array=config.get_array(sparse=False)))
        elif self.rng.rand() < self.random_fraction or (trials_budgets==self.all_budgets[-1]).sum() == 0:
            configs = self.space.sample_configuration(n_suggestions)
            for config in configs:
                trial_list.append(
                    Trial(configuration=config,
                          config_dict=config.get_dictionary(),
                          array=config.get_array(sparse=False)))
        else:
            # update target surrogate model
            X_all = np.asarray(self.trials.get_array())
            y_all = np.asarray(self.trials.get_history()[0])
            for budget in self.all_budgets:
                mask = trials_budgets == budget
                self.weighted_surrogate.train(
                    X_all[mask],
                    std_normalization(y_all[mask]),budget)
            # calculate base incuments (only use for acq base EI)
            std_incumbent_value = np.min(std_normalization(self.target_y[self.all_budgets[-1]]))

            self.acquisition_func.update(surrogate_model=self.weighted_surrogate,
                                         y_best=std_incumbent_value)
            # caculate weight for base+target model
            # weight = self.weighted_surrogate.get_weight(self.trials)
            # self.weighted_surrogate.update_weight(weight)
            # self.acquisition_func.update_weight(weight)
            # acq maximize
            configs = []
            configs = self.acq_maximizer.maximize(self.trials,
                                                  1000,
                                                  drop_self_duplicate=True,
                                                  _sorted=True)
            _idx = 0
            for n in range(n_suggestions):
                while _idx < len(configs):  # remove history suggest
                    if not self.trials.is_contain(configs[_idx]):
                        config = configs[_idx]
                        configs.append(config)
                        trial_list.append(
                            Trial(configuration=config,
                                  config_dict=config.get_dictionary(),
                                  array=config.get_array(sparse=False)))
                        _idx += 1

                        break
                    _idx += 1
                else:
                    assert False, "no more configs can be suggest"
                # surrogate = TST_surrogate(self.gps, self.target_model,
                #   self.similarity, self.rho)

        return trial_list

    def _observe(self, trial_list):
        # print(y)
        for trial in trial_list:
            self.trials.add_a_trial(trial)
            job_info = trial.info
            budget = job_info[Key.BUDGET]
            self.target_x[budget].append(trial.array)
            self.target_y[budget].append(trial.observe_value)

class MFES_BOHB_CG(BasicConfigGenerator):
    def __init__(self, sub_opt, cs, budget, max_pop_size, rng,
                 **kwargs) -> None:
        BasicConfigGenerator.__init__(self, cs, budget, max_pop_size, rng,
                                      **kwargs)
        self.sub_opt = sub_opt
        self.reset(max_pop_size)
    def _suggest(self, *args, **kwargs):
        return self.sub_opt._suggest(*args, **kwargs)

    def _observe(self, *args, **kwargs):
        return self.sub_opt._observe(*args, **kwargs)

    def update_weight(self, *args, **kwargs):
        return self.sub_opt.update_weight(*args, **kwargs)

alg_marker = 'mfes-bohb'
@alg_register.register(alg_marker)
class MFES_BOHB(BOHB):
    name = alg_marker

    def __init__(
            self,
            space: DenseConfigurationSpace,
            budget_bound=[9, 729],
            #  mutation_factor=0.5,
            #  crossover_prob=0.5,
            #  strategy='rand1_bin',
            eta: int = 3,
            seed: int = 42,
            round_limit: int = 1,
            bracket_limit=np.inf,
            boundary_fix_type='random',
            **kwargs):
        BOHB.__init__(self,
                      space,
                      budget_bound,
                      eta,
                      seed=seed,
                      round_limit=round_limit,
                      bracket_limit=bracket_limit,
                      boundary_fix_type=boundary_fix_type,
                      **kwargs)


        self.seed = seed
        self.logger = logger
        # Parameter for weight method `rank_loss_p_norm`.
        # self.power_num = power_num
        # Specify the weight learning method.
        # self.weight_srategy = weight_srategy
        self.weight_update_id = 0
        self.weight_changed_cnt = 0

        

        # self.incumbent_configs = []
        # self.incumbent_perfs = []

        # self.iterate_id = 0
        # self.iterate_r = []
        # self.hist_weights = list()

        # Saving evaluation statistics in Hyperband.
        # self.target_x = dict()
        # self.target_y = dict()
        # for index, item in enumerate(
        #         np.logspace(0, self.s_max, self.s_max + 1, base=self.eta)):
        #     r = int(item)
        #     self.iterate_r.append(r)
        #     self.target_x[r] = []
        #     self.target_y[r] = []

        # self.sls_max_steps = None
        # self.n_sls_iterations = 5
        # self.sls_n_steps_plateau_walk = 10



    def _init_subpop(self, **kwargs):
        """ List of DE objects corresponding to the budgets (fidelities)
        """
        self.cg = {}
        # types, bounds = get_types(self.space)

        sub_opt = SMBO(space=self.space,
                                all_budgets=self.budgets.tolist(),
                                weight_srategy='rank_loss_p_norm',
                                surrogate='rf',init_budget=0,
                                **kwargs)

        for i, b in enumerate(self._max_pop_size.keys()):
            self.cg[b] = MFES_BOHB_CG(sub_opt,
                                      self.space,
                                      seed=self.rng.randint(MAXINT),
                                      budget=b,
                                      max_pop_size=self._max_pop_size[b],
                                      rng=self.rng,
                                      **kwargs)

    def _acquire_candidate(self, bracket, budget):
        """ Generates/chooses a configuration based on the budget and iteration number
        """
        # select a parent/target
        # select a parent/target
        parent_id = self._get_next_idx_for_subpop(budget, bracket)

        if budget != bracket.budgets[0]:
            if bracket.is_new_rung():
                # TODO: check if generalizes to all budget spacings
                lower_budget, num_configs = bracket.get_lower_budget_promotions(
                    budget)
                self._get_promotion_candidate(lower_budget, budget,
                                              num_configs)
            # else: # 每一列中的第一行，随机生成config

        else:
            if bracket.is_new_rung():
                lower_budget, num_configs = bracket.get_lower_budget_promotions(
                    budget)
                self.cg[budget].population_fitness[:] = np.inf
                if self.round_recoder >= 1:
                    self.cg[budget].update_weight()

            # for b in reversed(self.budgets):
            #     if self.cg[b].kde_models:
            #         break
            # trial = self.cg[b]._suggest()[0]
            trials = self.cg[budget]._suggest(1)
            for i in range(len(trials)):
                self.cg[budget].population[parent_id + i] = trials[i].array
                # self.cg[budget].population[parent_id] = trial.array
        # parent_id = self._get_next_idx_for_subpop(budget, bracket)

        target = self.cg[budget].population[parent_id]
        # target = self.fix_boundary(target)
        return target, parent_id

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, True)
            fitness = trial.observe_value
            job_info = trial.info
            # learner_train_time = job_info.get(Key.EVAL_TIME, 0)
            budget = job_info[Key.BUDGET]
            parent_id = job_info['parent_id']
            individual = trial.array  # TODO
            for bracket in self.active_brackets:
                if bracket.bracket_id == job_info['bracket_id']:
                    # registering is IMPORTANT for Bracket Manager to perform SH
                    bracket.register_job(budget)  # may be new row
                    # bracket job complete
                    bracket.complete_job(
                        budget)  # IMPORTANT to perform synchronous SH
            self.cg[budget].population[parent_id] = individual
            self.cg[budget].population_fitness[parent_id] = fitness
            # updating incumbents
            if fitness < self.current_best_fitness:
                self.current_best = individual
                self.current_best_fitness = trial.observe_value
                self.current_best_trial = trial
        # for rung in range(bracket.n_rungs - 1, bracket.current_rung, -1):
        #     if self.cg[bracket.budgets[rung]].kde_models:
        #         break
        # else:
            self.cg[budget]._observe([trial])

        self._clean_inactive_brackets()

    def _sample_nonduplicate_config(self, num_configs=1):
        configs = list()
        sample_cnt = 0
        max_sample_cnt = 1000
        while len(configs) < num_configs:
            config = self.space.sample_configuration()[0]
            sample_cnt += 1
            if (not self.trials.is_contain(config)) and config not in configs:
                configs.append(config)
                sample_cnt = 0
                continue
            if sample_cnt >= max_sample_cnt:
                logger.warning(
                    'Cannot sample non duplicate configuration after %d iterations.'
                    % max_sample_cnt)
                configs.append(config)
                sample_cnt = 0
        return configs

opt_class = BOHB



import numpy as np
import time
from sklearn import svm, datasets
from sklearn.model_selection import cross_val_score
from ConfigSpace import ConfigurationSpace

from ConfigSpace.conditions import InCondition, LessThanCondition
from ConfigSpace.hyperparameters import \
    CategoricalHyperparameter, UniformFloatHyperparameter, UniformIntegerHyperparameter

from xbbo.utils.constants import MAXINT, Key


def mf_stochastic_count_one(config, info=None):
    '''
    $$
    f(x)=-\left(\sum_{x \in X_{c a t}} x+\sum_{x \in X_{c o n t}} b\left[\left(B_{p=x}\right)\right]\right)
    $$
    '''
    if info is None:
        info = {}
    budget = info.get(Key.BUDGET, 100)
    random_state = info.get('random_state', np.random.RandomState(0))
    xs = []
    rs = []
    st = time.time()
    for k in config:
        if k[0] == "x":
            xs.append(config[k])
        else:
            rs.append(random_state.binomial(1, config[k], size=int(budget)).mean())
    xs = -np.array(xs).sum()
    rs = -np.array(rs).sum()

    # result dict passed to DE/DEHB as function evaluation output
    res = {
        "fitness": xs+rs,  # must-have key that DE/DEHB minimizes
        # "cost": budget,  # must-have key that associates cost/runtime 
        Key.EVAL_TIME: time.time() - st
        # "info": dict() # optional key containing a dictionary of additional info
    }
    res.update(info)
    # dict representation that DEHB requires
    # res = {
    #     "fitness": loss,
    #     "cost": cost,
    #     "info": {"test_loss": test_loss, Key.BUDGET: budget}
    # }
    return res

def build_mf_SCO_space(rng, dim=8):
    cs = ConfigurationSpace(seed=rng.randint(MAXINT))
    xs = [CategoricalHyperparameter('x{}'.format(i), choices=[0,1]) for i in range(dim)]
    ys = [UniformFloatHyperparameter('y{}'.format(i), 0, 1, default_value=0.5) for i in range(dim)]
    cs.add_hyperparameters(xs)
    cs.add_hyperparameters(ys)
    return cs


def rosenbrock_2d(x):
    """ The 2 dimensional Rosenbrock function as a toy model
    The Rosenbrock function is well know in the optimization community and
    often serves as a toy problem. It can be defined for arbitrary
    dimensions. The minimium is always at x_i = 1 with a function value of
    zero. All input parameters are continuous. The search domain for
    all x's is the interval [-5, 10].
    """

    x1 = x["x1"]
    x2 = x["x2"]
    # x2 = x.get('x2', x1)

    val = 100. * (x2 - x1 ** 2.) ** 2. + (1 - x1) ** 2.
    return val

def branin(config):
    x1, x2 = config['x1'], config['x2']
    y = (x2 - 5.1 / (4 * np.pi ** 2) * x1 ** 2 + 5 / np.pi * x1 - 6) ** 2 \
        + 10 * (1 - 1 / (8 * np.pi)) * np.cos(x1) + 10
    return y

def rosenbrock_2d_hard(x):
    """ The 2 dimensional Rosenbrock function as a toy model
    The Rosenbrock function is well know in the optimization community and
    often serves as a toy problem. It can be defined for arbitrary
    dimensions. The minimium is always at x_i = 1 with a function value of
    zero. All input parameters are continuous. The search domain for
    all x's is the interval [-5, 10].
    """

    x1 = x["x1"]
    # x2 = x["x1"]
    x2 = x.get('x2', x1)
    x3 = x['x3']
    val = 100. * (x2 - x1 ** 2.) ** 2. + (1 - x1) ** 2.
    return val - (x3 == 2)

def build_space(rng):
    cs = ConfigurationSpace(seed=rng.randint(MAXINT))
    x0 = UniformFloatHyperparameter("x1", -5, 10, default_value=-3)
    x1 = UniformFloatHyperparameter("x2", -5, 10, default_value=-4)
    cs.add_hyperparameters([x0, x1])
    # con = LessThanCondition(x1, x0, 1.)
    # cs.add_condition(con)
    return cs

def build_branin_space(rng):
    cs = ConfigurationSpace(seed=rng.randint(MAXINT))
    x1 = UniformFloatHyperparameter("x1", -5, 10, default_value=0)
    x2 = UniformFloatHyperparameter("x2", 0, 15, default_value=0)
    cs.add_hyperparameters([x1, x2])
    return cs


def build_space_hard(rng):
    cs = ConfigurationSpace(seed=rng.randint(MAXINT))
    x0 = UniformFloatHyperparameter("x1", -5, 10, default_value=-3)
    x1 = UniformFloatHyperparameter("x2", -5, 10, default_value=-4)
    x2 = CategoricalHyperparameter("x3", choices=[0,1,2,3])
    cs.add_hyperparameters([x0, x1, x2])
    con = LessThanCondition(x1, x0, 1.)
    cs.add_condition(con)
    return cs


def zdt1(config):
    n = 30
    x1, x2 = config['x1'], config['x2']
    sigma = x2 # sum(input_x[1:])
    g = 1 + sigma * 9 / (n - 1)
    h = 1 - (x1 / g)**0.5
    return x1, g*h


def build_zdt1_space(rng):
    cs = ConfigurationSpace(seed=rng.randint(MAXINT))
    x0 = UniformFloatHyperparameter("x1", 0, 1)
    x1 = UniformFloatHyperparameter("x2", 0, 1)
    cs.add_hyperparameters([x0, x1])
    return cs

def svm_from_cfg(cfg):
    """ Creates a SVM based on a configuration and evaluates it on the
    iris-dataset using cross-validation. Note here random seed is fixed

    Parameters:
    -----------
    cfg: Configuration (ConfigSpace.ConfigurationSpace.Configuration)
        Configuration containing the parameters.
        Configurations are indexable!

    Returns:
    --------
    A crossvalidated mean score for the svm on the loaded data-set.
    """
    # For deactivated parameters, the configuration stores None-values.
    # This is not accepted by the SVM, so we remove them.
    cfg = {k: cfg[k] for k in cfg if cfg[k]}
    # And for gamma, we set it to a fixed value or to "auto" (if used)
    if "gamma" in cfg:
        cfg["gamma"] = cfg["gamma_value"] if cfg["gamma"] == "value" else "auto"
        cfg.pop("gamma_value", None)  # Remove "gamma_value"

    clf = svm.SVC(**cfg, random_state=42)

    scores = cross_val_score(clf, iris.data, iris.target, cv=5)
    return 1 - np.mean(scores)  # Minimize!

def build_svm_space(rng):
    global iris
    iris = datasets.load_iris()
        # Build Configuration Space which defines all parameters and their ranges
    cs = ConfigurationSpace(seed=rng.randint(MAXINT))

    # We define a few possible types of SVM-kernels and add them as "kernel" to our cs
    kernel = CategoricalHyperparameter("kernel",
                                       ["linear", "rbf", "poly", "sigmoid"],
                                       default_value="poly")
    cs.add_hyperparameter(kernel)

    # There are some hyperparameters shared by all kernels
    C = UniformFloatHyperparameter("C",
                                   0.001,
                                   1000.0,
                                   default_value=1.0,
                                   log=True)
    shrinking = CategoricalHyperparameter("shrinking", [True, False],
                                          default_value=True)
    cs.add_hyperparameters([C, shrinking])

    # Others are kernel-specific, so we can add conditions to limit the searchspace
    degree = UniformIntegerHyperparameter(
        "degree", 1, 5, default_value=3)  # Only used by kernel poly
    coef0 = UniformFloatHyperparameter("coef0", 0.0, 10.0,
                                       default_value=0.0)  # poly, sigmoid
    cs.add_hyperparameters([degree, coef0])

    use_degree = InCondition(child=degree, parent=kernel, values=["poly"])
    use_coef0 = InCondition(child=coef0,
                            parent=kernel,
                            values=["poly", "sigmoid"])
    cs.add_conditions([use_degree, use_coef0])

    # This also works for parameters that are a mix of categorical and values
    # from a range of numbers
    # For example, gamma can be either "auto" or a fixed float
    gamma = CategoricalHyperparameter(
        "gamma", ["auto", "value"],
        default_value="auto")  # only rbf, poly, sigmoid
    gamma_value = UniformFloatHyperparameter("gamma_value",
                                             0.0001,
                                             8,
                                             default_value=1,
                                             log=True)
    cs.add_hyperparameters([gamma, gamma_value])
    # We only activate gamma_value if gamma is set to "value"
    cs.add_condition(
        InCondition(child=gamma_value, parent=gamma, values=["value"]))
    # And again we can restrict the use of gamma in general to the choice of the kernel
    cs.add_condition(
        InCondition(child=gamma,
                    parent=kernel,
                    values=["rbf", "poly", "sigmoid"]))
    return cs


# License: 3-clause BSD
# Copyright (c) 2016-2018, Ml4AAD Group (http://www.ml4aad.org/)

import typing

from scipy.stats.qmc import LatinHypercube

from ConfigSpace.hyperparameters import Constant


from xbbo.configspace.space import DenseConfiguration
from xbbo.initial_design.base import InitialDesign
from xbbo.utils.constants import MAXINT

class LHDesign(InitialDesign):
    """Latin Hypercube design

    Attributes
    ----------
    configs : typing.List[Configuration]
        List of configurations to be evaluated
        Don't pass configs to the constructor;
        otherwise factorial design is overwritten
    """

    def _select_configurations(self, num=None) -> typing.List[DenseConfiguration]:
        """Selects a single configuration to run

        Returns
        -------
        config: Configuration
            initial incumbent configuration
        """
        design_num = num if num else self.init_budget
        dim = self.cs.get_dimensions(sparse=True)

        # constants = 0
        # for p in params:
        #     if isinstance(p, Constant):
        #         constants += 1

        lhd = LatinHypercube(d=dim, seed=self.rng.randint(0, MAXINT)).random(n=design_num)

        return self._transform_continuous_designs(design=lhd,
                                                  origin='LHD',
                                                  cs=self.cs)



from typing import List
import typing
from scipy import optimize, stats
import sklearn
# from sklearn.gaussian_process import kernels
from sklearn.gaussian_process.kernels import Kernel, KernelOperator
# from scipy.linalg import solve_triangular, cholesky
import numpy as np
# import GPy
from sklearn import gaussian_process

from xbbo.surrogate.base import BaseGP
from xbbo.surrogate.gp_kernels import HammingKernel, Matern, ConstantKernel, WhiteKernel
from xbbo.surrogate.gp_prior import HorseshoePrior, LognormalPrior, Prior, SoftTopHatPrior, TophatPrior
from xbbo.utils.util import get_types

VERY_SMALL_NUMBER = 1e-10


class GaussianTransform:
    """
    Transform data into Gaussian by applying psi = Phi^{-1} o F where F is the truncated ECDF.
    :param y: shape (n, dim)
    """
    def __init__(self, y: np.array):
        assert y.ndim == 2
        self.dim = y.shape[1]
        self.sorted = y.copy()
        self.sorted.sort(axis=0)

    @staticmethod
    def z_transform(series, values_sorted=None):
        # applies truncated ECDF then inverse Gaussian CDF.
        if values_sorted is None:
            values_sorted = sorted(series)

        def winsorized_delta(n):
            return 1.0 / (4.0 * n**0.25 * np.sqrt(np.pi * np.log(n)))

        delta = winsorized_delta(len(series))

        def quantile(values_sorted, values_to_insert, delta):
            res = np.searchsorted(values_sorted,
                                  values_to_insert) / len(values_sorted)
            return np.clip(res, a_min=delta, a_max=1 - delta)

        quantiles = quantile(values_sorted, series, delta)

        quantiles = np.clip(quantiles, a_min=delta, a_max=1 - delta)

        return stats.norm.ppf(quantiles)

    def transform(self, y: np.array):
        """
        :param y: shape (n, dim)
        :return: shape (n, dim), distributed along a normal
        """
        assert y.shape[1] == self.dim
        # compute truncated quantile, apply gaussian inv cdf
        return np.stack([
            self.z_transform(y[:, i], self.sorted[:, i])
            for i in range(self.dim)
        ]).T


class StandardTransform:
    def __init__(self, y: np.array):
        assert y.ndim == 2
        self.dim = y.shape[1]
        self.mean = y.mean(axis=0, keepdims=True)
        self.std = y.std(axis=0, keepdims=True)

    def transform(self, y: np.array):
        z = (y - self.mean) / np.clip(self.std, a_min=0.001, a_max=None)
        return z


class SEkernel():
    def __init__(self):
        self.initialize()

    def initialize(self):
        # self.sumF = 0.001
        # self.sumL = 0.001
        # self.sumY = 0.001
        self.sigma_f = 1
        self.sigma_l = 1  # TODO 之前设的是1
        self.sigma_y = 0.001

    def compute_kernel(self, x1, x2=None):
        if x2 is None:
            x2 = x1
            x2 = np.atleast_2d(x2)
            x1 = np.atleast_2d(x1)
            # noise = np.diag([self.sigma_y**2 for _ in range(x1.shape[0])])
            noise = np.eye(x1.shape[0]) * self.sigma_y**2
        else:
            x2 = np.atleast_2d(x2)
            x1 = np.atleast_2d(x1)
            noise = 0
        dist_matrix = np.sum(x1**2, 1).reshape(-1, 1) + np.sum(
            x2**2, 1) - 2 * (x1 @ x2.T)
        return self.sigma_f**2 * np.exp(
            -0.5 / self.sigma_l**2 * dist_matrix) + noise



class GPR_sklearn(BaseGP):
    def __init__(
        self,
        cs,
        #  min_sample=3,
        # alpha=0,
        rng=np.random.RandomState(0),
        n_opt_restarts: int = 10,
        instance_features: typing.Optional[np.ndarray] = None,
        pca_components: typing.Optional[int] = None,
        types=None,bounds=None,
        **kwargs
    ):
        if types is None or bounds is None:
            types, bounds = get_types(cs)
        # self.cached = {}

        super(GPR_sklearn, self).__init__(cs, types, bounds, rng,instance_features=instance_features,
            pca_components=pca_components,**kwargs)

        self.is_fited = False
        # self.alpha = alpha  # Fix RBF kernel error
        self.n_opt_restarts = n_opt_restarts
        self._n_ll_evals = 0
        self._set_has_conditions()

    def _get_kernel(self, ):
        cov_amp = ConstantKernel(
            2.0,
            constant_value_bounds=(np.exp(-10), np.exp(2)),
            prior=LognormalPrior(mean=0.0, sigma=1.0, rng=self.rng),
        )

        cont_dims = np.where(np.array(self.types) == 0)[0]
        cat_dims = np.where(np.array(self.types) != 0)[0]

        if len(cont_dims) > 0:
            exp_kernel = Matern(
                np.ones([len(cont_dims)]),
                [(np.exp(-6.754111155189306), np.exp(0.0858637988771976))
                 for _ in range(len(cont_dims))],
                nu=2.5,
                operate_on=cont_dims,
            )

        if len(cat_dims) > 0:
            ham_kernel = HammingKernel(
                np.ones([len(cat_dims)]),
                [(np.exp(-6.754111155189306), np.exp(0.0858637988771976))
                 for _ in range(len(cat_dims))],
                operate_on=cat_dims,
            )

        # assert (len(cont_dims) + len(cat_dims)) == len(
        #     scenario.cs.get_hyperparameters())

        noise_kernel = WhiteKernel(
            noise_level=1e-8,
            noise_level_bounds=(np.exp(-25), np.exp(2)),
            prior=HorseshoePrior(scale=0.1, rng=self.rng),
        )

        if len(cont_dims) > 0 and len(cat_dims) > 0:
            # both
            kernel = cov_amp * (exp_kernel * ham_kernel) + noise_kernel
        elif len(cont_dims) > 0 and len(cat_dims) == 0:
            # only cont
            kernel = cov_amp * exp_kernel + noise_kernel
        elif len(cont_dims) == 0 and len(cat_dims) > 0:
            # only cont
            kernel = cov_amp * ham_kernel + noise_kernel
        else:
            raise ValueError()
        # kernel = gaussian_process.kernels.ConstantKernel(
        #     constant_value=1  #, constant_value_bounds=(1e-4, 1e4)
        # ) * gaussian_process.kernels.RBF(
        #     length_scale=1  #, length_scale_bounds=(1e-4, 1e4)
        # )
        return kernel

    def _predict(self,
                X_test,
                cov_return_type: typing.Optional[str] = 'diagonal_cov'):
        '''
        return: \mu ,\sigma^2
        '''
        assert self.is_fited
        X_test = self._impute_inactive(X_test)
        if cov_return_type is None:
            mu = self.gp.predict(X_test)
            var = None

            if self.normalize_y:
                mu = self._untransform_y(mu)

        else:
            predict_kwargs = {'return_cov': False, 'return_std': True}
            if cov_return_type == 'full_cov':
                predict_kwargs = {'return_cov': True, 'return_std': False}

            mu, var = self.gp.predict(X_test, **predict_kwargs)

            if cov_return_type != 'full_cov':
                var = var**2  # since we get standard deviation for faster computation

            # Clip negative variances and set them to the smallest
            # positive float value
            var = np.clip(var, VERY_SMALL_NUMBER, np.inf)

            if self.normalize_y:
                mu, var = self._untransform_y(mu, var)

            if cov_return_type == 'diagonal_std':
                var = np.sqrt(
                    var)  # converting variance to std deviation if specified

        return mu, var

    def _get_gp(self) -> gaussian_process.GaussianProcessRegressor:
        return gaussian_process.GaussianProcessRegressor(
            kernel=self.kernel,
            normalize_y=False,
            optimizer=None,
            n_restarts_optimizer=
            -1,  # Do not use scikit-learn's optimization routine
            alpha=0,  # Governed by the kernel
            random_state=self.rng,
        )
    
    def _nll(self, theta: np.ndarray) -> typing.Tuple[float, np.ndarray]:
        """
        Returns the negative marginal log likelihood (+ the prior) for
        a hyperparameter configuration theta.
        (negative because we use scipy minimize for optimization)

        Parameters
        ----------
        theta : np.ndarray(H)
            Hyperparameter vector. Note that all hyperparameter are
            on a log scale.

        Returns
        ----------
        float
            lnlikelihood + prior
        """
        self._n_ll_evals += 1

        try:
            lml, grad = self.gp.log_marginal_likelihood(theta, eval_gradient=True)
        except np.linalg.LinAlgError:
            return 1e25, np.zeros(theta.shape)

        for dim, priors in enumerate(self._all_priors):
            for prior in priors:
                lml += prior.lnprob(theta[dim])
                grad[dim] += prior.gradient(theta[dim])

        # We add a minus here because scipy is minimizing
        if not np.isfinite(lml).all() or not np.all(np.isfinite(grad)):
            return 1e25, np.zeros(theta.shape)
        else:
            return -lml, -grad

    def _train(self, X: np.ndarray, y: np.ndarray, **kwargs):
        X = np.atleast_2d(X)
        X = self._impute_inactive(X)
        if self.normalize_y:
            y = self._normalize_y(y)
        if len(y.shape) == 1:
            self.n_objectives_ = 1
        else:
            self.n_objectives_ = y.shape[1]
        if self.n_objectives_ == 1:
            y = y.flatten()

        n_tries = 10
        for i in range(n_tries):
            try:
                self.gp = self._get_gp()  # new model
                self.gp.fit(X, y)
                break
            except np.linalg.LinAlgError as e:
                if i == n_tries:
                    raise e
                # Assume that the last entry of theta is the noise
                theta = np.exp(self.kernel.theta)
                theta[-1] += 1
                self.kernel.theta = np.log(theta)
        if self.do_optimize:
            self._all_priors = self._get_all_priors(add_bound_priors=False)
            self.hypers = self._optimize()
            self.gp.kernel.theta = self.hypers
            self.gp.fit(X, y)
        else:
            self.hypers = self.gp.kernel.theta
        self.is_fited = True

    # def _get_all_priors(
    #     self,
    #     add_bound_priors: bool = True,
    #     add_soft_bounds: bool = False,
    # ) -> List[List[Prior]]:
    #     # Obtain a list of all priors for each tunable hyperparameter of the kernel
    #     all_priors = []
    #     to_visit = []
    #     to_visit.append(self.gp.kernel.k1)
    #     to_visit.append(self.gp.kernel.k2)
    #     while len(to_visit) > 0:
    #         current_param = to_visit.pop(0)
    #         if isinstance(current_param, KernelOperator):
    #             to_visit.insert(0, current_param.k1)
    #             to_visit.insert(1, current_param.k2)
    #             continue
    #         elif isinstance(current_param, Kernel):
    #             hps = current_param.hyperparameters
    #             assert len(hps) == 1
    #             hp = hps[0]
    #             if hp.fixed:
    #                 continue
    #             bounds = hps[0].bounds
    #             for i in range(hps[0].n_elements):
    #                 priors_for_hp = []
    #                 if current_param.prior is not None:
    #                     priors_for_hp.append(current_param.prior)
    #                 if add_bound_priors:
    #                     if add_soft_bounds:
    #                         priors_for_hp.append(
    #                             SoftTopHatPrior(
    #                                 lower_bound=bounds[i][0],
    #                                 upper_bound=bounds[i][1],
    #                                 rng=self.rng,
    #                                 exponent=2,
    #                             ))
    #                     else:
    #                         priors_for_hp.append(
    #                             TophatPrior(
    #                                 lower_bound=bounds[i][0],
    #                                 upper_bound=bounds[i][1],
    #                                 rng=self.rng,
    #                             ))
    #                 all_priors.append(priors_for_hp)
    #     return all_priors

    def _optimize(self) -> np.ndarray:
        """
        Optimizes the marginal log likelihood and returns the best found
        hyperparameter configuration theta.

        Returns
        -------
        theta : np.ndarray(H)
            Hyperparameter vector that maximizes the marginal log likelihood
        """

        log_bounds = [(b[0], b[1]) for b in self.gp.kernel.bounds]

        # Start optimization from the previous hyperparameter configuration
        p0 = [self.gp.kernel.theta]
        if self.n_opt_restarts > 0:
            dim_samples = []

            prior = None  # type: typing.Optional[typing.Union[typing.List[Prior], Prior]]
            for dim, hp_bound in enumerate(log_bounds):
                prior = self._all_priors[dim]
                # Always sample from the first prior
                if isinstance(prior, list):
                    if len(prior) == 0:
                        prior = None
                    else:
                        prior = prior[0]
                prior = typing.cast(typing.Optional[Prior], prior)
                if prior is None:
                    try:
                        sample = self.rng.uniform(
                            low=hp_bound[0],
                            high=hp_bound[1],
                            size=(self.n_opt_restarts, ),
                        )
                    except OverflowError:
                        raise ValueError(
                            'OverflowError while sampling from (%f, %f)' %
                            (hp_bound[0], hp_bound[1]))
                    dim_samples.append(sample.flatten())
                else:
                    dim_samples.append(
                        prior.sample_from_prior(self.n_opt_restarts).flatten())
            p0 += list(np.vstack(dim_samples).transpose())

        theta_star = None
        f_opt_star = np.inf
        for i, start_point in enumerate(p0):
            theta, f_opt, _ = optimize.fmin_l_bfgs_b(self._nll,
                                                     start_point,
                                                     bounds=log_bounds)
            if f_opt < f_opt_star:
                f_opt_star = f_opt
                theta_star = theta
        return theta_star

    # def _set_has_conditions(self) -> None:
    #     has_conditions = len(self.configspace.get_conditions()) > 0
    #     to_visit = []
    #     to_visit.append(self.kernel)
    #     while len(to_visit) > 0:
    #         current_param = to_visit.pop(0)
    #         if isinstance(current_param,
    #                       sklearn.gaussian_process.kernels.KernelOperator):
    #             to_visit.insert(0, current_param.k1)
    #             to_visit.insert(1, current_param.k2)
    #             current_param.has_conditions = has_conditions
    #         elif isinstance(current_param,
    #                         sklearn.gaussian_process.kernels.Kernel):
    #             current_param.has_conditions = has_conditions
    #         else:
    #             raise ValueError(current_param)





# License: 3-clause BSD
# Copyright (c) 2016-2018, Ml4AAD Group (http://www.ml4aad.org/)

from typing import List

from xbbo.initial_design.base import InitialDesign
from xbbo.configspace.space import DenseConfiguration

class RandomDesign(InitialDesign):
    """Initial design that evaluates random configurations."""

    def _select_configurations(self, num=None) -> List[DenseConfiguration]:
        """Select a random configuration.

        Returns
        -------
        config: Configuration()
            Initial incumbent configuration
        """
        design_num = num if num else self.init_budget
        configs = self.cs.sample_configuration(size=design_num)
        if design_num == 1:
            configs = [configs]
        for config in configs:
            config.origin = 'Random initial design.'
        return configs



import importlib, os

from xbbo.core.register import Register

alg_register = Register("all avaliable search algorithms")

for model in os.listdir(os.path.dirname(__file__)):
    if model.endswith('.py'):
        module = importlib.import_module('.' + model[:-3], __package__)
for model in os.listdir(os.path.dirname(__file__)+'/multi_fidelity'):
    if model.endswith('.py'):
        module = importlib.import_module('.' + model[:-3], __package__+'.multi_fidelity')

# __all__ = alg_register.keys()
# def get_opt_class(opt_name):

#     if opt_name == "hyperopt_optimizer":
#         from xbbo.search_algorithm import hyperopt_optimizer as opt
#     elif opt_name == "nevergrad_optimizer":
#         from xbbo.search_algorithm import nevergrad_optimizer as opt
#     elif opt_name == "opentuner_optimizer":
#         from xbbo.search_algorithm import opentuner_optimizer as opt
#     elif opt_name == "pysot_optimizer":
#         from xbbo.search_algorithm import pysot_optimizer as opt
#     elif opt_name == "random_optimizer":
#         from xbbo.search_algorithm import random_optimizer as opt
#     elif opt_name == "scikit_optimizer":
#         from xbbo.search_algorithm import scikit_optimizer as opt
#     elif opt_name == "bore_optimizer":
#         from xbbo.search_algorithm import bore_optimizer as opt
#     elif opt_name == "sng_optimizer":
#         from xbbo.search_algorithm import sng_optimizer as opt
#     elif opt_name == "toy_tpe_optimizer":
#         from xbbo.search_algorithm import toy_tpe_optimizer as opt
#     elif opt_name == "cem_optimizer":
#         from xbbo.search_algorithm import cem_optimizer as opt
#     elif opt_name == "anneal_optimizer":
#         from xbbo.search_algorithm import anneal_optimizer as opt
#     elif opt_name == "anneal_tree_optimizer":
#         from xbbo.search_algorithm import anneal_tree_optimizer as opt
#     elif opt_name == "transfer_tst_optimizer":
#         from xbbo.search_algorithm import transfer_tst_optimizer as opt
#     elif opt_name == "transfer_taf_optimizer":
#         from xbbo.search_algorithm import transfer_taf_optimizer as opt
#     elif opt_name == "transfer_baseline_optimizer":
#         from xbbo.search_algorithm import transfer_baseline_optimizer as opt
#     elif opt_name == "transfer_rgpe_optimizer":
#         from xbbo.search_algorithm import _transfer_rgpe_optimizer as opt
#     elif opt_name == "transfer_rs_optimizer":
#         from xbbo.search_algorithm import transfer_rs_optimizer as opt
#     elif opt_name == "transfer_baseline_optimizer_":
#         from xbbo.search_algorithm import transfer_baseline_optimizer_ as opt
#     elif opt_name == "transfer_tst_optimizer_":
#         from xbbo.search_algorithm import transfer_tst_optimizer_ as opt
#     elif opt_name == "transfer_taf_optimizer_":
#         from xbbo.search_algorithm import transfer_taf_optimizer as opt
#     elif opt_name == "transfer_rgpe_mean_optimizer_":
#         from xbbo.search_algorithm import transfer_rgpe_mean_optimizer as opt
#     elif opt_name == "transfer_taf_rgpe_optimizer_":
#         from xbbo.search_algorithm import transfer_taf_rgpe_optimizer as opt
#     elif opt_name == "transfer_RMoGP_optimizer_":
#         from xbbo.search_algorithm import transfer_RMoGP_optimizer as opt
#     elif opt_name == "de_optimizer":
#         from xbbo.search_algorithm import de_optimizer as opt
#     elif opt_name == "cma_optimizer":
#         from xbbo.search_algorithm import cma_optimizer as opt
#     elif opt_name == "nsga_optimizer":
#         from xbbo.search_algorithm import nsga_optimizer as opt
#     elif opt_name == "regularizedEA_optimizer":
#         from xbbo.search_algorithm import regularizedEA_optimizer as opt
#     elif opt_name == "pbt_optimizer":
#         from xbbo.search_algorithm import pbt_optimizer as opt
#     elif opt_name == "turbo_optimizer":
#         from xbbo.search_algorithm import turbo_optimizer as opt
#     elif opt_name == "bo_optimizer":
#         from xbbo.search_algorithm import bo_optimizer as opt
#     else:
#         assert False, f"{opt_name} is not in xbbo/search_algorithm or not change xbbo/search_algorithm/__init__.py"

#     return opt.opt_class


import typing
import numpy as np
from xbbo.surrogate.gaussian_process import GPR_sklearn
from xbbo.surrogate.transfer.base_surrogate import BaseModel
VERY_SMALL_NUMBER = 1e-10


class TST_surrogate(GPR_sklearn):
    def __init__(
            self,
            cs,
            #  min_sample=3,
            # alpha=0,
            base_models: typing.List[BaseModel],
            rng=np.random.RandomState(0),
            n_opt_restarts: int = 10,
            instance_features: typing.Optional[np.ndarray] = None,
            pca_components: typing.Optional[int] = None,
            **kwargs):
        super().__init__(cs,
                         rng,
                         n_opt_restarts,
                         instance_features=instance_features,
                         pca_components=pca_components,
                         **kwargs)
        self.base_models = base_models
        
    def target_model_predict(self, X: np.ndarray, cov_return_type: typing.Optional[str] = 'diagonal_cov') -> typing.Tuple[np.ndarray, np.ndarray]:
        return super()._predict(X, cov_return_type=cov_return_type)

    def update_weight(self, w, rho=None):
        assert w.min() >= 0
        if rho:
            self.w = np.array(w, dtype='float')
            self.selfWeight = rho
        else:
            self.selfWeight = w[-1]
            self.w = np.array(w[:-1], dtype='float')
        #non_zero_weight_indices = (self.w**2 > 0).nonzero()[0]
        non_zero_weight_indices = self.w.nonzero()[0]
        self.pre_weight_model = []
        for idx in non_zero_weight_indices:
            self.pre_weight_model.append(self.base_models[idx])
        self.weight = self.w[non_zero_weight_indices]
        norm_coeff = (self.weight.sum() + self.selfWeight)
        self.weight /= norm_coeff
        self.selfWeight /= norm_coeff

    def _predict(self,
                 X_test,
                 cov_return_type: typing.Optional[str] = 'diagonal_cov'):
        '''
        return: \mu ,\sigma^2
        '''
        assert self.is_fited
        # denominator = self.selfWeight

        X_test = self._impute_inactive(X_test)
        if cov_return_type is None:
            mu = self.gp.predict(X_test)
            var = None
            mu *= self.selfWeight
            for d in range(len(self.weight)):
                mu += self.weight[d] * self.pre_weight_model[
                    d]._predict_normalize(X_test, cov_return_type=None)[0]
                # denominator += self.weight[d]
            # mu /= denominator
            if self.normalize_y:
                mu = self._untransform_y(mu)

        else:
            predict_kwargs = {'return_cov': False, 'return_std': True}
            if cov_return_type == 'full_cov':
                predict_kwargs = {'return_cov': True, 'return_std': False}

            mu, var = self.gp.predict(X_test, **predict_kwargs)
            mu *= self.selfWeight
            for d in range(len(self.weight)):
                mu += self.weight[d] * self.pre_weight_model[
                    d]._predict_normalize(X_test, cov_return_type=None)[0]
                # denominator += self.weight[d]
            # mu /= denominator

            if cov_return_type != 'full_cov':
                var = var**2  # since we get standard deviation for faster computation

            # Clip negative variances and set them to the smallest
            # positive float value
            var = np.clip(var, VERY_SMALL_NUMBER, np.inf)

            if self.normalize_y:
                mu, var = self._untransform_y(mu, var)

            if cov_return_type == 'diagonal_std':
                var = np.sqrt(
                    var)  # converting variance to std deviation if specified

        return mu, var

    # def kendallTauCorrelation(self, d, x, y):
    #     '''
    #     计算第d个datasets与new datasets的 相关性
    #     (x, y) 为newdatasets上的history结果
    #     '''
    #     if y is None or len(y) < 2:
    #         return self.rho
    #     disordered_pairs = total_pairs = 0
    #     for i in range(len(y)):
    #         for j in range(i+1, len(y)): # FIXME
    #             if (y[i] < y[j] != self.gps[d].cached_predict(
    #                     x[i]) < self.gps[d].cached_predict(x[j])):
    #                 disordered_pairs += 1
    #             total_pairs += 1
    #     t = disordered_pairs / total_pairs / self.bandwidth
    #     return self.rho * (1 - t * t) if t < 1 else 0



# License: MIT
# ref: https://github.com/thomas-young-2013/open-box/blob/master/openbox/surrogate/skrf.py
import logging
import typing
import numpy as np
from typing import List, Optional, Tuple, Union

from xbbo.surrogate.base import BaseRF
from xbbo.configspace.space import DenseConfigurationSpace
from xbbo.utils.constants import MAXINT
from xbbo.utils.util import get_types


logger = logging.getLogger(__name__)


class RandomForestSurrogate(BaseRF):

    def __init__(
            self,
            configspace: DenseConfigurationSpace,
            ensemble_size: int = 10,
            normalize_y: bool = True,
            instance_features: typing.Optional[np.ndarray] = None,
            pca_components: typing.Optional[int] = None,
            rng: np.random.RandomState = np.random.RandomState(42),types=None,bounds=None,
            **kwargs
    ):
    
        self.model_config = dict()
        self.model_config["n_estimators"] = 10
        self.model_config["criterion"] = "mse"
        self.model_config["max_depth"] = 12
        self.model_config["min_samples_split"] = 3
        self.model_config["min_samples_leaf"] = 3
        self.model_config["min_weight_fraction_leaf"] = 0.
        self.model_config["max_features"] = 5. / 6.
        self.model_config["max_leaf_nodes"] = None
        self.model_config["n_jobs"] = -1
        # self.model_config["random_state"] = -1
        # self.model_config["max_samples"] = 1.

        self.ensemble_size = ensemble_size
        self.models = list()
        self.configspace = configspace
        self.rng = rng
        self.random_seeds = self.rng.randint(low=1, high=MAXINT, size=self.ensemble_size)
        if types is None or bounds is None:
            types, bounds = get_types(configspace)
        super().__init__(
            configspace=configspace,
            types=types,
            bounds=bounds,
            instance_features=instance_features,
            pca_components=pca_components,
            **kwargs
        )

        self.normalize_y = normalize_y
        self.is_trained = False

    def _train(self, X: np.ndarray, y: np.ndarray, **kwargs):
        """
        Train a Random Forest Regression model on X and y

        Parameters
        ----------
        X: np.ndarray (N, D)
            Input data points. The dimensionality of X is (N, D),
            with N as the number of points and D is the number of features.
        y: np.ndarray (N,)
            The corresponding target values.
        """
        from sklearn.ensemble import RandomForestRegressor
        X = self._impute_inactive(X)
        if self.normalize_y:
            y = self._normalize_y(y)

        self.models = list()
        for i in range(self.ensemble_size):
            configs = self.model_config.copy()
            configs["random_state"] = self.random_seeds[i]
            rf_model = RandomForestRegressor(**configs)
            rf_model.fit(X, y)
            self.models.append(rf_model)

        self.is_trained = True
        return self

    def _predict(self, X_test: np.ndarray, **kwargs):
        r"""
        Returns the predictive mean and variance of the objective function

        Parameters
        ----------
        X_test: np.ndarray (N, D)
            Input test points

        Returns
        ----------
        np.array(N,)
            predictive mean
        np.array(N,)
            predictive variance

        """
        if not self.is_trained:
            raise Exception('Model has to be trained first!')

        X_test = self._impute_inactive(X_test)

        predictions = list()
        for i, model in enumerate(self.models):
            pred = model.predict(X_test)
            # print(i)
            predictions.append(pred)

        m = np.mean(predictions, axis=0)
        v = np.var(predictions, axis=0)

        # Clip negative variances and set them to the smallest
        # positive float value
        if v.shape[0] == 1:
            v = np.clip(v, np.finfo(v.dtype).eps, np.inf)
        else:
            v = np.clip(v, np.finfo(v.dtype).eps, np.inf)
            v[np.where((v < np.finfo(v.dtype).eps) & (v > -np.finfo(v.dtype).eps))] = 0

        if self.normalize_y:
            m, v = self._untransform_y(m, v)

        return m, v

    def _normalize_y(self, y: np.ndarray) -> np.ndarray:
        """Normalize data to zero mean unit standard deviation.
        """
        self.mean_y_ = np.mean(y)
        self.std_y_ = np.std(y)
        if self.std_y_ == 0:
            self.std_y_ = 1
        return (y - self.mean_y_) / self.std_y_

    def _untransform_y(
            self,
            y: np.ndarray,
            var: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Transform zeromean unit standard deviation data into the regular space.
        """
        y = y * self.std_y_ + self.mean_y_
        if var is not None:
            var = var * self.std_y_ ** 2
            return y, var
        return y

    def _impute_inactive(self, X: np.ndarray) -> np.ndarray:
        X = X.copy()
        X[~np.isfinite(X)] = -1
        return X



# License: 3-clause BSD
# Copyright (c) 2016-2018, Ml4AAD Group (http://www.ml4aad.org/)

import itertools

from ConfigSpace.hyperparameters import Constant, NumericalHyperparameter, \
    CategoricalHyperparameter, OrdinalHyperparameter
import numpy as np


from xbbo.configspace.space import DenseConfiguration, deactivate_inactive_hyperparameters
from xbbo.initial_design.base import InitialDesign


class FactorialInitialDesign(InitialDesign):
    """Factorial initial design

    Attributes
    ----------
    configs : typing.List[Configuration]
        List of configurations to be evaluated
        Don't pass configs to the constructor;
        otherwise factorial design is overwritten
    """
    def _select_configurations(self, num=None) -> DenseConfiguration:
        """Selects a single configuration to run

        Returns
        -------
        config: Configuration
            initial incumbent configuration
        """
        params = self.cs.get_hyperparameters()

        values = []
        mid = []
        for param in params:
            if isinstance(param, Constant):
                v = [param.value]
                mid.append(param.value)
            elif isinstance(param, NumericalHyperparameter):
                v = [param.lower, param.upper]
                mid.append(np.average([param.lower, param.upper]))
            elif isinstance(param, CategoricalHyperparameter):
                v = list(param.choices)
                mid.append(param.choices[0])
            elif isinstance(param, OrdinalHyperparameter):
                v = [param.sequence[0], param.sequence[-1]]
                length = len(param.sequence)
                mid.append(param.sequence[int(length / 2)])
            values.append(v)

        factorial_design = itertools.product(*values)

        self.logger.debug("Initial Design")
        configs = [self.cs.get_default_configuration()]
        # add middle point in space
        conf_dict = dict([(p.name, v) for p, v in zip(params, mid)])
        middle_conf = deactivate_inactive_hyperparameters(conf_dict, self.cs)
        configs.append(middle_conf)

        # add corner points
        for design in factorial_design:
            conf_dict = dict([(p.name, v) for p, v in zip(params, design)])
            conf = deactivate_inactive_hyperparameters(conf_dict, self.cs)
            conf.origin = "Factorial Design"
            configs.append(conf)
            self.logger.debug(conf)
        size = len(configs)
        self.logger.debug("Size of factorial design: %d" % (size))
        self.init_budget = size
        return configs



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import optim
from torch.autograd import Variable
from torch.nn.parameter import Parameter
from torchvision import datasets, transforms
from torch.utils.data import DataLoader, Dataset

from xbbo.utils.constants import MAXINT

from xbbo.core import TestFunction

class LossIsNaN(Exception):
    pass




class Model(TestFunction):

    def __init__(self, cfg, seed, **kwargs):
        # np.random.seed(cfg.GENERAL.random_seed)
        self.cfg = cfg
        # self.dim = 30
        # assert self.dim % 2 == 0
        super().__init__(seed=seed)

        self.api_config = self._load_api_config()
        torch.seed(self.rng.randint(MAXINT))
        torch.manual_seed(self.rng.randint(MAXINT))
        self.device = torch.device(kwargs.get('device', 'cpu'))

        self.theta = Parameter(torch.FloatTensor([0.9, 0.9]).to(self.device))
        # self.opt_wrap = lambda params: optim.SGD(self.net.parameters(), lr=lr, momentum=momentum)
        self.opt = optim.SGD([self.theta], lr=0.01)
        self.step_num = 0
        self.history_hp = [] # for record strategy
        self.trajectory_hp = []
        self.trajectory_loss = [] # 记录该个体score过程
        self.history_loss = [] # 记录使用了（考虑权重迁移）hp-stategy后的score过程
        self.hp = torch.empty(2, device=self.device)
        self.obj_val_func = lambda theta: 1.2 - (theta ** 2).sum()
        self.obj_train_func = lambda theta, h: 1.2 - ((h * theta) ** 2).sum()

        self.trajectory_theta = []

    def __len__(self): # one epoch has how many batchs
        return 1

    def update_hp(self, params: dict):
        self.history_hp.append((self.step_num, params)) # 在该steps上更改超参，acc为该step时的结果（受该step*前*所有超参影响）
        self.trajectory_hp.append((self.step_num, params))
        self.trajectory_theta.append(self.theta.detach().cpu().numpy())
        self.hp[0] = params['h1']
        self.hp[1] = params['h2']

    def step(self, num): # train need training(optimizer)
        for it in range(num):
            self.trajectory_theta.append(self.theta.detach().cpu().numpy())
            loss = self.obj_train_func(self.theta, self.hp)
            if np.isnan(loss.item()):
                print("Loss is NaN.")
                self.step_num += 1
                return
                # raise LossIsNaN
            self.opt.zero_grad()
            loss.backward()
            self.opt.step()
            self.step_num += 1


    def evaluate(self): # val no training need(optimizer)
        with torch.no_grad():
            loss = self.obj_val_func(self.theta).item()
        self.loss = np.inf if np.isnan(loss) else loss
        self.trajectory_loss.append((self.step_num, self.loss))
        self.history_loss.append((self.step_num, self.loss))
        return self.loss

    def load_checkpoint(self, checkpoint):
        with torch.no_grad():
            self.theta.set_(checkpoint['model_state_dict'])
        # self.opt.load_state_dict(checkpoint['optim_state_dict'])

    def save_checkpoint(self):
        checkpoint = dict(model_state_dict=self.theta.data.clone())
        return checkpoint

    def _load_api_config(self):
        return {
            'h1': {
                'type': 'float', 'warp': 'linear', 'range': [0, 1]},
            'h2': {
                'type': 'float', 'warp': 'linear', 'range': [0, 1]
            }
        }




import math
from typing import Optional, List, Tuple, cast

import numpy as np

from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trial, Trials
from . import alg_register


@alg_register.register('nsga2')
class NSGAII(AbstractOptimizer):
    '''
    reference: https://zhuanlan.zhihu.com/p/144807879
    '''
    def __init__(self,
                 space: DenseConfigurationSpace,
                 seed: int = 42,
                 llambda=None,
                 **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   **kwargs)
        self.dimension = self.space.get_dimensions()
        self.bounds = self.space.get_bounds()

        self.num_of_tour_particips = kwargs.get('n_part',2)
        self.llambda = llambda if llambda else 4 + math.floor(
            3 * math.log(self.dimension))  # (eq. 48)
        self.population = self.create_initial_population()
        self.population_y = None
        self.mu = kwargs.get('mu',20) # 交叉和变异算法的分布指数
        self.mum = kwargs.get('mum',20)
        self.crossrate = kwargs.get('crossrate',0.9)
        # ---

        self.trials = Trials(space,dim=self.dimension)
        self.cur = 0
        self.gen = 0
        self.listy = []

    def _suggest(self, n_suggestions=1):
        assert self.llambda % n_suggestions == 0
        trial_list = []
        for n in range(n_suggestions):
            new_individual = self.population[self.cur]
            new_individual = np.clip(new_individual, self.bounds.lb,
                                     self.bounds.ub)
            # array = self.feature_to_array(new_individual)
            config = DenseConfiguration.from_array(
                self.space, new_individual)
            self.cur += 1
            trial_list.append(
                Trial(config,
                      config_dict=config.get_dictionary(),
                      array=new_individual,
                      origin='NSGAII',
                      loc=self.cur))

        # self._num_suggestions += n_suggestions
        return trial_list

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, permit_duplicate=True)
            self.listy.append(trial.observe_value)
        if self.cur == len(self.population):
            if self.population_y is None:
                self.population_y = np.asarray(self.listy)
            else:
                self.population_y = np.concatenate(
                    [self.population_y, self.listy], axis=0)

            # remove dead
            # self.population_y = np.asarray(self.population_y[-self.llambda:])
            # self.population = np.asarray(self.population[-self.llambda:])
            ranks_, crowding_distance = self.fast_nondominated_sort(
                self.population_y)
            s_id = sorted(
                zip(ranks_, -crowding_distance, range(len(self.population_y))))
            s_id = np.array([s[-1] for s in s_id])

            self.population = np.delete(self.population,
                                        s_id[self.llambda:],
                                        axis=0)
            self.population_y = np.delete(self.population_y,
                                          s_id[self.llambda:],
                                          axis=0)

            new_s_id = s_id[:self.llambda]
            # tmp = np.argsort(new_s_id)
            # new_s_id[tmp] = np.arange(self.llambda)
            # s_id = new_s_id
            rank = np.argsort(new_s_id)
            # FIXME
            parents_id = list(self.selection_tournament(
                rank, self.llambda / 2))

            self.children = self.create_children(parents_id)

            # ---
            self.population = self.population[parents_id]
            self.population_y = self.population_y[parents_id]
            self.population = np.append(self.population, self.children, axis=0)
            self.cur = len(self.population_y)
            assert self.cur < len(self.population)

            self.listy = []
            self.gen += 1

    def selection_tournament(self, s_id, num):
        # parents = []
        parents_id = set()
        while len(parents_id) < num:
            parent_id = self.__tournament(s_id)
            # if parent_id in parents_id:
            #     continue
            parents_id.add(parent_id)
            # parents.append(self.population[parent_id])
        return parents_id

    def fast_nondominated_sort(self, points):
        if len(points.shape) < 2:
            points = points[..., None]
        individual_num = len(points)
        ranks = np.zeros(individual_num)
        crowding_distance = np.zeros(individual_num)
        r = 0
        c = individual_num
        m = np.min(points, axis=0)
        M = np.max(points, axis=0)
        # while c > 0:
        #     extended = np.tile(points, (points.shape[0], 1, 1))
        #     dominance = np.sum(np.logical_and(
        #         np.all(extended <= np.swapaxes(extended, 0, 1), axis=2),
        #         np.any(extended < np.swapaxes(extended, 0, 1), axis=2)), axis=1)
        #
        #     points[dominance == 0] = 1e9  # mark as used
        #     ranks[dominance == 0] = r
        #     r += 1
        #     c -= np.sum(dominance == 0)
        extended = np.tile(points, (points.shape[0], 1, 1))
        dominance = np.sum(np.logical_and(
            np.all(extended <= np.swapaxes(extended, 0, 1), axis=2),
            np.any(extended < np.swapaxes(extended, 0, 1), axis=2)),
                           axis=1)
        _num = sorted(set(dominance))
        while c > 0:
            idx_r = np.where(dominance == _num[r])[0]
            assert len(idx_r) > 0
            crowding_distance[idx_r] = self.calculate_crowding_distance(
                points[idx_r], m, M)
            ranks[idx_r] = r
            c -= len(idx_r)
            r += 1
        return ranks, crowding_distance

    def calculate_crowding_distance(self, points, m, M):
        cd = np.zeros(len(points))
        if len(cd) <= points.shape[-1]:
            return cd
        for n in range(points.shape[-1]):
            s_id = np.argsort(points[:, n])
            cd[s_id[[0, -1]]] = [np.inf, np.inf]
            for idx in range(len(s_id[1:-1])):
                cd[s_id[idx]] += (points[s_id[idx + 1], n] -
                                  points[s_id[idx - 1], n]) / (M[n] - m[n])
        return cd

    def create_initial_population(self):
        # return self.rng.rand(self.llambda, self.dimension)
        return self.rng.uniform(0,
                                1,
                                size=(self.llambda, self.dimension))

    def create_children(self, parents_id):
        children = []
        num = len(parents_id)
        list(range(num))
        for i in range(num):
            if self.rng.rand() > self.crossrate:
                child = self.__mutate(
                    self.population[self.rng.choice(parents_id)])
                children.append(child)
            else:
                parent1_id, parent2_id = self.rng.choice(parents_id,
                                                         2,
                                                         replace=False)
                parent1 = np.copy(self.population[parent1_id])
                parent2 = np.copy(self.population[parent2_id])
                child1, child2 = self.__crossover(parent1, parent2)

                children.append(child1)
                children.append(child2)

        return children

    def __crossover(self, individual1, individual2):
        u = self.rng.rand(len(individual1))
        bq = np.where(u < 0.5, 2 * u**(1 / (self.mu + 1)),
                      (1 / (2 * (1 - u)))**(1 / (self.mu + 1)))
        child1 = 0.5 * (((1 + bq) * individual1) + (1 - bq) * individual2)

        child2 = 0.5 * (((1 - bq) * individual1) + (1 + bq) * individual2)

        # child1 = individual1.copy()
        # child2 = individual2.copy()
        # i = self.rng.randint(child1.shape[1])
        # t = child1[i]
        # child1[i] = child2[i]
        # child2[i] = t
        return np.clip(child1, self.bounds.lb, self.bounds.ub), np.clip(child2, self.bounds.lb, self.bounds.ub)

    def __mutate(self, parent):
        child = parent.copy()
        r = self.rng.rand(len(child))
        delta = np.where(r < 0.5, (2 * r)**(1 / (self.mum + 1)) - 1,
                         1 - (2 * (1 - r))**(1 / (self.mum + 1)))
        return np.clip(child + delta, self.bounds.lb, self.bounds.ub)

    def __tournament(self, rank):
        participants = self.rng.choice(len(self.population),
                                       self.num_of_tour_particips,
                                       replace=False)

        return participants[np.argsort(rank[participants])[0]]


opt_class = NSGAII



# from collections import OrderedDict
import numpy as np

from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
# from xbbo.core.stochastic import Category, Uniform
from xbbo.core.trials import Trial, Trials
from . import alg_register
from xbbo.initial_design import ALL_avaliable_design


@alg_register.register('anneal')
class Anneal(AbstractOptimizer):
    def __init__(self,
                 space,
                 seed: int = 42,
                 suggest_limit=10,
                 initial_design: str = 'sobol',
                 **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   suggest_limit=suggest_limit,
                                   **kwargs)
        self.dimension = self.space.get_dimensions()

        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, ta_run_limit=suggest_limit, **kwargs)
        self.init_budget = self.initial_design.init_budget
        self.initial_design_configs = self.initial_design.select_configurations(
        )

        self.avg_best_idx = kwargs.get("avg_best_idx", 2.0)
        self.shrink_coef = kwargs.get("shrink_coef", 0.1)
        self.cat_nodes = []
        self.num_nodes = []
        for k in self.space.map:
            if k == 'const':
                continue
            if k != 'num':
                self.cat_nodes.extend([
                    {
                        "idx": self.space.map[k].src[i],
                        # "activate_num": 0,
                        "n_choice": self.space.map[k].sizes[i]
                    } for i in range(len(self.space.map[k].src))
                ])
            else:
                self.num_nodes.extend([{
                    "idx": self.space.map[k].src[i],
                } for i in range(len(self.space.map[k].src))])

        self.trials = Trials(space,dim=self.dimension)

    def _suggest(self, n_suggestions=1):
        trial_list = []
        if (self.trials.trials_num) < self.init_budget:
            assert self.trials.trials_num % n_suggestions == 0
            configs = self.initial_design_configs[
                int(n_suggestions *
                    self.trials.trials_num):int(n_suggestions *
                                                (self.trials.trials_num + 1))]
            for config in configs:
                trial_list.append(
                    Trial(configuration=config,
                          config_dict=config.get_dictionary(),
                          array=config.get_array()))
        else:
            for n in range(n_suggestions):
                X = self.trials.get_array()
                Y = np.asarray(self.trials._his_observe_value)

                array = np.empty(self.dimension)
                for node in self.cat_nodes:
                    idx = node["idx"]
                    x_ = X[:, idx].astype(
                        'int')  # deactivate conditional variable
                    mask = ~np.isnan(x_)
                    x = x_[mask]
                    y = Y[mask]
                    best_val = self._choose_best_trail_feature(x, y)
                    X[:, idx]
                    p = self._handle_category(best_val, len(y),
                                              node["n_choice"])
                    array[idx] = self.rng.choice(node["n_choice"], p=p)
                for node in self.num_nodes:
                    idx = node["idx"]
                    x_ = X[:, idx]  # deactivate conditional variable
                    mask = ~np.isnan(x_)
                    x = x_[mask]
                    y = Y[mask]
                    best_val = self._choose_best_trail_feature(x, y)
                    # X[:, idx]
                    low, high = self._handle_uniform(best_val, len(y))
                    array[idx] = self.rng.uniform(low, high)
                config = DenseConfiguration.from_array(self.space, array)
                array = config.get_array()
                trial_list.append(
                    Trial(configuration=config,
                          config_dict=config.get_dictionary(),
                          array=array))
                # valid_indices = np.argwhere(~np.isnan(sparse_array)).ravel()
                # for node in self.cat_nodes:
                #     if node["idx"] in valid_indices:
                #         node["activate_num"] += 1
                # for node in self.cat_nodes:
                #     if node["idx"] in valid_indices:
                #         node["activate_num"] += 1

        return trial_list

    def _shrinking(self, activate_num):
        T = activate_num
        return 1.0 / (1.0 + T * self.shrink_coef)

    def _handle_category(self, val, activate_num, n_choice):
        prior = self._shrinking(activate_num)
        if val is None:
            return np.ones(n_choice) / n_choice
        val1 = np.atleast_1d(val)
        counts = np.bincount(val1, minlength=n_choice) / val1.size
        return (1 - prior) * counts + prior * 1 / n_choice

    def _handle_uniform(self, midpt, activate_num, low=0, high=1):
        if midpt is None:
            return low, high
        width = (high - low) * self._shrinking(activate_num)

        half = 0.5 * width
        min_midpt = low + half  # 避免超出边界
        max_midpt = high - half
        clipped_midpt = np.clip(midpt, min_midpt, max_midpt)
        return clipped_midpt - half, clipped_midpt + half

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial)

    def _choose_best_trail_feature(self, his_x, his_y):
        if len(his_y) <= self.avg_best_idx:
            return None
        good_idx = self.rng.geometric(1.0 / self.avg_best_idx) - 1
        good_idx = np.clip(good_idx, 0,
                           self.trials.trials_num - 1).astype("int32")
        picks = np.argsort(his_y)[good_idx]
        return his_x[picks]


opt_class = Anneal


from yacs.config import CfgNode
import os, argparse, sys, time
import hashlib
import shutil
from contextlib import redirect_stdout

_C = CfgNode()

cfg = _C

_C.GENERAL = CfgNode()
_C.GENERAL.gpu = ''
_C.GENERAL.random_seed = 42
_C.GENERAL.exp_dir_root = './exp'
_C.GENERAL.pipeline = 'BBO'

_C.BBO = CfgNode()
_C.BBO.DATASETS = CfgNode()
_C.BBO.DATASETS.name = ''
_C.BBO.DATASETS.path = ''
_C.BBO.DATASETS.batch_size = 128

_C.OPTM = CfgNode()
_C.OPTM.name = 'rs' 
_C.OPTM.n_suggestions = 1
_C.OPTM.n_obj = 1 
_C.OPTM.max_call = 30
_C.OPTM.pop_size = 0 # for PBT
_C.OPTM.epoch = 30 # for PBT
_C.OPTM.interval = 0.5 # for PBT
# _C.OPTM.fraction = 0.2 # for PBT

_C.OPTM.kwargs = CfgNode(new_allowed=True)

_C.NAS = CfgNode() # TODO

_C.TEST_PROBLEM = CfgNode()
_C.TEST_PROBLEM.name = 'toy-problems' # filename
_C.TEST_PROBLEM.kwargs = CfgNode(new_allowed=True)
# _C.TEST_PROBLEM.func_evals = ('raw', 'noise') # 放非优化器优化目标的结果、metrics
# _C.TEST_PROBLEM.losses = ('val', 'test') # 必须前n_obj个为opt的优化目标
# _C.TEST_PROBLEM.metrics = ()
# _C.TEST_PROBLEM.kwargs.func_name = 'rosenbrock'

# _C.TEST_PROBLEM.kwargs = CfgNode()
_C.TEST_PROBLEM.kwargs.func_name = 'rosenbrock' # if empty, TestProblem must provide `_load_api_config` method
# _C.TEST_PROBLEM.kwargs.hp = CfgNode(new_allowed=True)
_C.TEST_PROBLEM.kwargs.dim = 30

# _C.SPACE = CfgNode()
# _C.SPACE.NAME = 'darts'
# # Loss function
# _C.SPACE.LOSS_FUN = "cross_entropy"
# # num of classes
# _C.SPACE.NUM_CLASSES = 10
# # Init channel
# _C.SPACE.CHANNEL = 16
# # number of layers
# _C.SPACE.LAYERS = 8
# # number of nodes in a cell
# _C.SPACE.NODES = 4
# # number of  PRIMITIVE
# _C.SPACE.PRIMITIVES = []
# # number of nodes in a cell
# _C.SPACE.BASIC_OP = []

def load_cfg_fom_args(cfg_, description = "Config file options.", argv=None):
    """Load config from command line arguments and set any specified options.
       How to use: python xx.py --cfg path_to_your_config.cfg test1 0 test2 True
       opts will return a list with ['test1', '0', 'test2', 'True'], yacs will compile to corresponding values
    """

    parser = argparse.ArgumentParser(description=description)
    help_s = "Config file location"
    parser.add_argument('-c, '"--cfg", dest="cfg_file",
                        help=help_s, required=True, type=str)
    parser.add_argument('-r, '"--repeat", dest="repeat_num",
                        help=help_s, required=True, type=int)
    parser.add_argument("--mark", dest="mark_label", default='',
                        help=help_s, required=False, type=str)
    help_s = "See pycls/core/config.py for all options"
    parser.add_argument("opts", help=help_s, default=None,
                        nargs=argparse.REMAINDER)
    # if len(sys.argv) == 1:
    #     parser.print_help()
    #     sys.exit(1)
    args = parser.parse_args(argv)
    cfg_.merge_from_file(args.cfg_file)
    cfg_.merge_from_list(args.opts)
    cfg_.repeat_num = args.repeat_num
    cfg_.mark_label = args.mark_label
    if not os.path.exists(cfg_.GENERAL.exp_dir_root):
        os.mkdir(cfg_.GENERAL.exp_dir_root)
    m = hashlib.md5(cfg_.__repr__().encode('utf-8'))
    exp_dir = time.strftime('/%Y-%m-%d__%H_%M_%S__',time.localtime(time.time()))+m.hexdigest()
    # _C.runtime = time.time()
    if cfg_.mark_label == '':
        cfg_.mark_label = '{}-{}'.format(cfg_.OPTM.name, m.hexdigest())
    cfg_.GENERAL.exp_dir = cfg_.GENERAL.exp_dir_root + exp_dir # TODO
    if os.path.exists(cfg_.GENERAL.exp_dir):
        assert False
    os.mkdir(cfg_.GENERAL.exp_dir)
    os.mkdir(cfg_.GENERAL.exp_dir + '/res')
    os.mkdir(cfg_.GENERAL.exp_dir + '/log')
    # os.mkdir(_C.GENARAL.exp_dir+'/script')
    project_dir = os.path.abspath(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '../..'
        )
    )
    shutil.copytree(project_dir +'/xbbo/', cfg_.GENERAL.exp_dir + '/scripts/')


    # _C.freeze()
    # cfg_.freeze()



    with open(os.path.join(cfg_.GENERAL.exp_dir, os.path.basename(args.cfg_file)), 'w') as f:
        with redirect_stdout(f): print(cfg_.dump())


if __name__ == '__main__':
    # print(cfg.BBO)
    print(1 if _C.NAS else 0)
    print(1 if _C.OPTM.kwargs else 0)


import numpy as np
from time import time
import tqdm

from xbbo.search_space import problem_register
from xbbo.search_algorithm import alg_register
from xbbo.configspace import build_space
from xbbo.utils.constants import MAXINT
from xbbo.utils.record import Record


class BBO:

    def __init__(self, cfg, seed):
        # setup TestProblem
        self.cfg = cfg
        self.rng = np.random.RandomState(seed)
        self.function_instance = problem_register[cfg.TEST_PROBLEM.name](seed=self.rng.randint(MAXINT), **cfg.TEST_PROBLEM.kwargs)

        self.api_config = self.function_instance.get_api_config()  # 优化的hp
        self.config_spaces = build_space(self.api_config,seed=self.rng.randint(MAXINT))

        # Setup optimizer
        opt_class = alg_register[cfg.OPTM.name]
        self.optimizer_instance = opt_class(self.config_spaces,suggest_limit=cfg.OPTM.max_call,seed=self.rng.randint(MAXINT), **dict(cfg.OPTM.kwargs))


        self.n_suggestions = cfg.OPTM.n_suggestions
        self.n_obj = cfg.OPTM.n_obj

        assert self.n_suggestions >= 1, "batch size must be at least 1"
        assert self.n_obj >= 1, "Must be at least one objective"


        # self.suggest_time = np.zeros(n_calls)
        # self.observe_time = np.zeros(n_calls)
        # self.eval_time = np.zeros((n_calls, n_suggestions))
        # self.function_evals = np.zeros((n_calls, n_suggestions, self.n_obj))
        # self.suggest_log = [None] * n_calls
        self.n_calls = cfg.OPTM.max_call
        self.record = Record(self.cfg.GENERAL.exp_dir)


    def evaluate(self, param):
        return self.function_instance.evaluate(param)

    def run(self):
        pbar = tqdm.tqdm(range(self.n_calls))
        pbar.set_description(f"Optimizer {self.cfg.OPTM.name} is running:")
        
        for ii in pbar:

            tt = time()
            trial_list = self.optimizer_instance.suggest(self.n_suggestions)  # TODO 1

            # try:
            #     next_points, features = self.optimizer_instance.suggest(self.n_suggestions)  # TODO 1
            # except Exception as e:
            #     # logger.warning("Failure in optimizer suggest. Falling back to random search.")
            #     # logger.exception(e, exc_info=True)
            #     print(json.dumps({"optimizer_suggest_exception": {'iter': ii}}))
            #     # api_config = self.function_instance.get_api_config()
            #     # TODO 直接随机采样
            #     x_guess_configs = self.optimizer_instance.space.sample_configuration(size=self.n_suggestions) # a list
            #     next_points = [x_guess_config.get_dict_unwarped() for x_guess_config in x_guess_configs]
            #     features = [x_guess_config.get_array(sparse=False) for x_guess_config in x_guess_configs]

            suggest_time = time() - tt

            assert len(trial_list) == self.n_suggestions, "invalid number of suggestions provided by the optimizer"
            # eval_time = [None for _ in range(self.n_suggestions)]
            function_evals = []
            # losses = []
            for trial in (trial_list):
                # try:
                tt = time()

                f_current_eval = self.evaluate(trial.config_dict) # TODO 2
                eval_time = time() - tt

                # except Exception as e:
                #     f_current_eval = np.full((len(self.cfg.TEST_PROBLEM.func_evals),), np.inf, dtype=float)
                #     loss = np.full((len(self.cfg.TEST_PROBLEM.losses),), np.inf, dtype=float)


                trial.add_observe_value(observe_value=f_current_eval, obs_info={Key.EVAL_TIME:eval_time})
                function_evals.append(f_current_eval)
            # if self.cfg.OPTM.n_obj == 1:
            #     eval_list = np.asarray(function_evals)[:, :self.cfg.OPTM.n_obj].ravel().tolist() # TODO
            # else:
            #     raise NotImplementedError()
                # eval_list = np.array(losses)[:, :self.cfg.OPTM.n_obj].tolist() # TODO
            # assert self.cfg.OPTM.n_obj == 1
            tt = time()
            self.optimizer_instance.observe(trial_list)  # TODO 3
            # try:
            #     self.optimizer_instance.observe(features, eval_list)  # TODO 3
            # except Exception as e:
                # logger.warning("Failure in optimizer observe. Ignoring these observations.")
                # logger.exception(e, exc_info=True)
                # print(json.dumps({"optimizer_observe_exception": {'iter': ii}}))
            observe_time = time() - tt
            timing = {
                         'suggest_time_per_suggest': suggest_time,
                         'observe_time_per_suggest': observe_time,
                         'eval_time_per_suggest': sum(trial.time for trial in trial_list)
            }
            self.record.append([trial.dense_array if trial.sparse_array is None else trial.sparse_array for trial in trial_list], function_evals, timing=timing, suggest_point=[trial.config_dict for trial in trial_list])
            # print(self.optimizer_instance.trials.best_observe_value)
            print(function_evals)

        print(self.optimizer_instance.trials.best_observe_value)


class BBO_REBAR(BBO):
    def __init__(self):
        BBO.__init__(self)
    
    def evaluate(self,params):
        return BBO.evaluate(self, params)
    
    def run(self):
        pass